# -*- coding: utf-8 -*-

"""Converter for FlyBase Genes."""

import logging
from typing import Iterable, Mapping, Set

import pandas as pd
from tqdm.auto import tqdm

from pyobo import Reference
from pyobo.struct import Obo, Term, from_species, orthologous
from pyobo.utils.io import multisetdict
from pyobo.utils.path import ensure_df

__all__ = [
    "FlyBaseGetter",
]

logger = logging.getLogger(__name__)

BASE_URL = "http://ftp.flybase.net/releases"
PREFIX = "flybase"
NAME = "FlyBase"


class FlyBaseGetter(Obo):
    """An ontology representation of FlyBase's gene nomenclature."""

    ontology = bioversions_key = PREFIX
    typedefs = [from_species, orthologous]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(force=force, version=self._version_or_raise)


def _get_names(version: str, force: bool = False) -> pd.DataFrame:
    url = f"{BASE_URL}/FB{version}/precomputed_files/genes/fbgn_fbtr_fbpp_expanded_fb_{version}.tsv.gz"
    df = ensure_df(
        PREFIX,
        url=url,
        force=force,
        version=version,
        skiprows=4,
        usecols=[0, 1, 2, 3, 4],
        skipfooter=1,
        engine="python",
    )
    return df


def _get_organisms(version: str, force: bool = False) -> Mapping[str, str]:
    """Get mapping from abbreviation column to NCBI taxonomy ID column."""
    url = f"http://ftp.flybase.net/releases/FB{version}/precomputed_files/species/organism_list_fb_{version}.tsv.gz"
    df = ensure_df(
        PREFIX, url=url, force=force, version=version, skiprows=4, header=None, usecols=[2, 4]
    )
    df.dropna(inplace=True)
    return dict(df.values)


def _get_definitions(version: str, force: bool = False) -> Mapping[str, str]:
    url = f"http://ftp.flybase.net/releases/FB{version}/precomputed_files/genes/automated_gene_summaries.tsv.gz"
    df = ensure_df(
        PREFIX, url=url, force=force, version=version, skiprows=2, header=None, usecols=[0, 1]
    )
    return dict(df.values)


def _get_human_orthologs(version: str, force: bool = False) -> Mapping[str, Set[str]]:
    url = (
        f"http://ftp.flybase.net/releases/FB{version}/precomputed_files/"
        f"orthologs/dmel_human_orthologs_disease_fb_{version}.tsv.gz"
    )
    df = ensure_df(
        PREFIX,
        url=url,
        force=force,
        version=version,
        skiprows=2,
        header=None,
        usecols=[0, 2],
        names=["flybase_id", "hgnc_id"],
    )
    return multisetdict(df.values)


def _get_synonyms(version, force):
    url = f"http://ftp.flybase.net/releases/FB{version}/precomputed_files/synonyms/fb_synonym_fb_{version}.tsv.gz"
    df = ensure_df(PREFIX, url=url, force=force, version=version, skiprows=4, usecols=[0, 2])
    return df  # TODO use this


def get_obo(force: bool = False) -> Obo:
    """Get OBO."""
    return FlyBaseGetter(force=force)


GTYPE_TO_SO = {
    "SRP_RNA_gene": "0001269",
    "protein_coding_gene": "0001217",
    "pseudogene": "0000336",
    "lncRNA_gene": "0002127",
    "snRNA_gene": "0001268",
    "antisense_lncRNA_gene": "0002182",
    "tRNA_gene": "0001272",
    "rRNA_gene": "0001637",
    "snoRNA_gene": "0001267",
    "RNase_P_RNA_gene": "0001639",
    "rRNA_5S_gene": "0002238",
    "ncRNA_gene": "0001263",
    "RNase_MRP_RNA_gene": "0001640",
    "rRNA_18S_gene": "0002236",
    "rRNA_5_8S_gene": "0002240",
    "miRNA_gene": "0001265",
    "rRNA_28S_gene": "0002239",
    "mt_LSU_rRNA_gene": None,
}


def get_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get terms."""
    definitions = _get_definitions(version=version, force=force)
    abbr_to_taxonomy = _get_organisms(version=version, force=force)
    names_df = _get_names(version=version, force=force)
    human_orthologs = _get_human_orthologs(version=version, force=force)
    missing_taxonomies = set()

    so = {}
    for gtype in names_df[names_df.columns[1]].unique():
        so_id = GTYPE_TO_SO.get(gtype)
        if so_id is None:
            logger.warning(
                "FlyBase gene type is missing mapping to Sequence Ontology (SO): %s", gtype
            )
        else:
            so[gtype] = Reference.auto("SO", so_id)

    for _, reference in sorted(so.items()):
        yield Term(reference=reference)
    for organism, gtype, identifier, symbol, name in tqdm(
        names_df.values, unit_scale=True, unit="gene", desc="FlyBase genes"
    ):
        term = Term.from_triple(
            prefix=PREFIX,
            identifier=identifier,
            name=symbol if pd.notna(symbol) else None,
            definition=definitions.get(identifier),
        )
        if gtype and pd.notna(gtype) and gtype in so:
            term.append_parent(so[gtype])
        if pd.notna(name):
            term.append_synonym(name)
        for hgnc_curie in human_orthologs.get(identifier, []):
            if not hgnc_curie or pd.isna(hgnc_curie):
                continue
            hgnc_ortholog = Reference.from_curie(hgnc_curie, auto=True)
            if hgnc_ortholog is None:
                tqdm.write(f"[{PREFIX}] {identifier} had invalid ortholog: {hgnc_curie}")
            else:
                term.append_relationship(orthologous, hgnc_ortholog)
        taxonomy_id = abbr_to_taxonomy.get(organism)
        if taxonomy_id is not None:
            term.set_species(taxonomy_id)
        elif organism not in missing_taxonomies:
            tqdm.write(f"[{PREFIX}] missing mapping for species abbreviation: {organism}")
            missing_taxonomies.add(organism)
        yield term

    if missing_taxonomies:
        tqdm.write(f"[{PREFIX}] there were {len(missing_taxonomies)} missing taxa in flybase genes")


if __name__ == "__main__":
    FlyBaseGetter.cli()



# -*- coding: utf-8 -*-

"""Import NCIT mappings."""

from typing import Iterable

import pandas as pd

from ...constants import (
    PROVENANCE,
    SOURCE_ID,
    SOURCE_PREFIX,
    TARGET_ID,
    TARGET_PREFIX,
    XREF_COLUMNS,
)
from ...utils.path import ensure_df

__all__ = [
    "iter_ncit_dfs",
    "get_ncit_go_df",
    "get_ncit_chebi_df",
    "get_ncit_hgnc_df",
    "get_ncit_uniprot_df",
]

PREFIX = "ncit"

HGNC_MAPPINGS_URL = (
    "https://ncit.nci.nih.gov/ncitbrowser/ajax?action="
    + "export_mapping&dictionary=NCIt_to_HGNC_Mapping&version=1.0"
)

GO_MAPPINGS_URL = (
    "https://ncit.nci.nih.gov/ncitbrowser/ajax?action="
    + "export_mapping&dictionary=GO_to_NCIt_Mapping&version=1.1"
)

CHEBI_MAPPINGS_URL = (
    "https://ncit.nci.nih.gov/ncitbrowser/ajax?action="
    + "export_mapping&dictionary=NCIt_to_ChEBI_Mapping&version=1.0"
)

# url_swissprot = 'https://ncit.nci.nih.gov/ncitbrowser/ajax?action=' \
#                 'export_mapping&uri=https://evs.nci.nih.gov/ftp1/' \
#                 'NCI_Thesaurus/Mappings/NCIt-SwissProt_Mapping.txt'

UNIPROT_MAPPINGS_URL = (
    "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/Mappings/NCIt-SwissProt_Mapping.txt"
)


def get_ncit_xrefs_df() -> pd.DataFrame:
    """Get all NCIT mappings in a single dataframe."""
    return pd.concat(iter_ncit_dfs())


def iter_ncit_dfs() -> Iterable[pd.DataFrame]:
    """Iterate all NCIT mappings dataframes."""
    yield get_ncit_hgnc_df()
    yield get_ncit_chebi_df()
    yield get_ncit_uniprot_df()
    yield get_ncit_go_df()


def get_ncit_hgnc_df() -> pd.DataFrame:
    """Get NCIT-HGNC mappings.

    In this file, the only association type was mapsTo.
    """
    df = ensure_df(
        PREFIX,
        url=HGNC_MAPPINGS_URL,
        name="ncit_hgnc.csv",
        sep=",",
        usecols=["Source Code", "Target Code"],
    )
    df.rename(columns={"Source Code": SOURCE_ID, "Target Code": TARGET_ID}, inplace=True)
    df[TARGET_ID] = df[TARGET_ID].map(lambda s: s[len("HGNC:") :])
    df.dropna(inplace=True)

    df[SOURCE_PREFIX] = "ncit"
    df[TARGET_PREFIX] = "hgnc"
    df[PROVENANCE] = HGNC_MAPPINGS_URL
    df = df[XREF_COLUMNS]
    return df


def get_ncit_go_df() -> pd.DataFrame:
    """Get NCIT-GO mappings.

    In this file, the only association type was mapsTo.
    """
    df = ensure_df(PREFIX, url=GO_MAPPINGS_URL, name="ncit_go.csv", sep=",")
    # The data is flipped here
    df.rename(columns={"Source Code": TARGET_ID, "Target Code": SOURCE_ID}, inplace=True)
    df[TARGET_ID] = df[TARGET_ID].map(lambda s: s[len("GO:")])
    df.dropna(inplace=True)

    df[SOURCE_PREFIX] = "ncit"
    df[TARGET_PREFIX] = "go"
    df[PROVENANCE] = GO_MAPPINGS_URL
    df = df[XREF_COLUMNS]
    return df


def get_ncit_chebi_df() -> pd.DataFrame:
    """Get NCIT-ChEBI mappings.

    In this file, the only association type was mapsTo.
    """
    df = ensure_df(PREFIX, url=CHEBI_MAPPINGS_URL, name="ncit_chebi.csv", sep=",")
    df.rename(columns={"Source Code": SOURCE_ID, "Target Code": TARGET_ID}, inplace=True)
    df[TARGET_ID] = df[TARGET_ID].map(lambda s: s[len("CHEBI:")])
    df.dropna(inplace=True)

    df[SOURCE_PREFIX] = "ncit"
    df[TARGET_PREFIX] = "chebi"
    df[PROVENANCE] = CHEBI_MAPPINGS_URL
    df = df[XREF_COLUMNS]
    return df


def get_ncit_uniprot_df() -> pd.DataFrame:
    """Get NCIT-UniProt mappings.

    In this file, the only association type was mapsTo.
    """
    df = ensure_df(PREFIX, url=UNIPROT_MAPPINGS_URL, name="ncit_uniprot.csv")
    df.rename(columns={"NCIt Code": SOURCE_ID, "SwissProt ID": TARGET_ID}, inplace=True)
    df[SOURCE_PREFIX] = "ncit"
    df[TARGET_PREFIX] = "uniprot"
    df[PROVENANCE] = UNIPROT_MAPPINGS_URL
    df = df[XREF_COLUMNS]
    return df



# -*- coding: utf-8 -*-

"""High-level API for nomenclature."""

import logging
import subprocess
import zipfile
from functools import lru_cache
from typing import Callable, List, Mapping, Optional, Set, TypeVar

from .alts import get_primary_identifier
from .utils import get_version
from ..getters import NoBuild, get_ontology
from ..identifier_utils import normalize_curie, wrap_norm_prefix
from ..utils.cache import cached_collection, cached_mapping, cached_multidict
from ..utils.path import prefix_cache_join

__all__ = [
    "get_name",
    "get_name_by_curie",
    "get_ids",
    "get_id_name_mapping",
    "get_name_id_mapping",
    "get_definition",
    "get_id_definition_mapping",
    "get_synonyms",
    "get_id_synonyms_mapping",
]

logger = logging.getLogger(__name__)


def get_name_by_curie(curie: str) -> Optional[str]:
    """Get the name for a CURIE, if possible."""
    prefix, identifier = normalize_curie(curie)
    if prefix and identifier:
        return get_name(prefix, identifier)
    return None


X = TypeVar("X")

NO_BUILD_PREFIXES = set()


def _help_get(
    f: Callable[[str], Mapping[str, X]],
    prefix: str,
    identifier: str,
    force: bool = False,
    strict: bool = False,
    version: Optional[str] = None,
) -> Optional[X]:
    """Get the result for an entity based on a mapping maker function ``f``."""
    try:
        mapping = f(prefix, force=force, strict=strict, version=version)  # type:ignore
    except NoBuild:
        if prefix not in NO_BUILD_PREFIXES:
            logger.warning("[%s] unable to look up results with %s", prefix, f)
            NO_BUILD_PREFIXES.add(prefix)
        return None

    if not mapping:
        if prefix not in NO_BUILD_PREFIXES:
            logger.warning("[%s] no results produced with %s", prefix, f)
            NO_BUILD_PREFIXES.add(prefix)
        return None

    primary_id = get_primary_identifier(prefix, identifier)
    return mapping.get(primary_id)


@wrap_norm_prefix
def get_name(prefix: str, identifier: str) -> Optional[str]:
    """Get the name for an entity."""
    return _help_get(get_id_name_mapping, prefix, identifier)


@lru_cache()
@wrap_norm_prefix
def get_ids(
    prefix: str, force: bool = False, strict: bool = False, version: Optional[str] = None
) -> Set[str]:
    """Get the set of identifiers for this prefix."""
    if prefix == "ncbigene":
        from ..sources.ncbigene import get_ncbigene_ids

        logger.info("[%s] loading name mappings", prefix)
        rv = get_ncbigene_ids()
        logger.info("[%s] done loading name mappings", prefix)
        return rv

    if version is None:
        version = get_version(prefix)
    path = prefix_cache_join(prefix, name="ids.tsv", version=version)

    @cached_collection(path=path, force=force)
    def _get_ids() -> Set[str]:
        if force:
            logger.info("[%s v%s] forcing reload for names", prefix, version)
        else:
            logger.debug(
                "[%s v%s] no cached identifiers found. getting from OBO loader", prefix, version
            )
        ontology = get_ontology(prefix, force=force, strict=strict, version=version)
        return ontology.get_ids()

    return set(_get_ids())


@lru_cache()
@wrap_norm_prefix
def get_id_name_mapping(
    prefix: str, *, force: bool = False, strict: bool = False, version: Optional[str] = None
) -> Mapping[str, str]:
    """Get an identifier to name mapping for the OBO file."""
    if prefix == "ncbigene":
        from ..sources.ncbigene import get_ncbigene_id_to_name_mapping

        logger.info("[%s] loading name mappings", prefix)
        rv = get_ncbigene_id_to_name_mapping()
        logger.info("[%s] done loading name mappings", prefix)
        return rv

    if version is None:
        version = get_version(prefix)
    path = prefix_cache_join(prefix, name="names.tsv", version=version)

    @cached_mapping(path=path, header=[f"{prefix}_id", "name"], force=force)
    def _get_id_name_mapping() -> Mapping[str, str]:
        if force:
            logger.debug("[%s v%s] forcing reload for names", prefix, version)
        else:
            logger.debug("[%s v%s] no cached names found. getting from OBO loader", prefix, version)
        ontology = get_ontology(prefix, force=force, strict=strict, version=version)
        return ontology.get_id_name_mapping()

    try:
        return _get_id_name_mapping()
    except (zipfile.BadZipFile, subprocess.CalledProcessError):
        logger.exception("[%s v%s] could not load", prefix, version)
        return {}


@lru_cache()
@wrap_norm_prefix
def get_name_id_mapping(prefix: str, force: bool = False) -> Mapping[str, str]:
    """Get a name to identifier mapping for the OBO file."""
    id_name = get_id_name_mapping(prefix=prefix, force=force)
    return {v: k for k, v in id_name.items()}


@wrap_norm_prefix
def get_definition(prefix: str, identifier: str) -> Optional[str]:
    """Get the definition for an entity."""
    return _help_get(get_id_definition_mapping, prefix, identifier)


def get_id_definition_mapping(
    prefix: str,
    *,
    force: bool = False,
    strict: bool = False,
    version: Optional[str] = None,
) -> Mapping[str, str]:
    """Get a mapping of descriptions."""
    if version is None:
        version = get_version(prefix)
    path = prefix_cache_join(prefix, name="definitions.tsv", version=version)

    @cached_mapping(path=path, header=[f"{prefix}_id", "definition"], force=force)
    def _get_mapping() -> Mapping[str, str]:
        logger.info(
            "[%s v%s] no cached descriptions found. getting from OBO loader", prefix, version
        )
        ontology = get_ontology(prefix, force=force, strict=strict, version=version)
        return ontology.get_id_definition_mapping()

    return _get_mapping()


@wrap_norm_prefix
def get_synonyms(prefix: str, identifier: str) -> Optional[List[str]]:
    """Get the synonyms for an entity."""
    return _help_get(get_id_synonyms_mapping, prefix, identifier)


@wrap_norm_prefix
def get_id_synonyms_mapping(
    prefix: str,
    *,
    force: bool = False,
    strict: bool = False,
    version: Optional[str] = None,
) -> Mapping[str, List[str]]:
    """Get the OBO file and output a synonym dictionary."""
    if version is None:
        version = get_version(prefix)
    path = prefix_cache_join(prefix, name="synonyms.tsv", version=version)

    @cached_multidict(path=path, header=[f"{prefix}_id", "synonym"], force=force)
    def _get_multidict() -> Mapping[str, List[str]]:
        logger.info("[%s v%s] no cached synonyms found. getting from OBO loader", prefix, version)
        ontology = get_ontology(prefix, force=force, strict=strict, version=version)
        return ontology.get_id_synonyms_mapping()

    return _get_multidict()



# -*- coding: utf-8 -*-

"""Converter for PubChem Compound."""

import logging
from typing import Iterable, Mapping, Optional

import bioversions
import pandas as pd
from bioregistry.utils import removeprefix
from tqdm.auto import tqdm

from ..api import get_name_id_mapping
from ..struct import Obo, Reference, Synonym, Term
from ..utils.iter import iterate_gzips_together
from ..utils.path import ensure_df, ensure_path

__all__ = [
    "PubChemCompoundGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "pubchem.compound"


def _get_pubchem_extras_url(version: Optional[str], end: str) -> str:
    if version is None:
        version = bioversions.get_version("pubchem")
    return f"ftp://ftp.ncbi.nlm.nih.gov/pubchem/Compound/Monthly/{version}/Extras/{end}"


class PubChemCompoundGetter(Obo):
    """An ontology representation of the PubChem Compound database."""

    ontology = PREFIX
    bioversions_key = "pubchem"

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get PubChem Compound OBO."""
    return PubChemCompoundGetter(force=force)


def _get_cid_smiles_df(version: str) -> pd.DataFrame:
    url = _get_pubchem_extras_url(version, "CID-SMILES.gz")
    return ensure_df(PREFIX, url=url, version=version, dtype=str)


def get_pubchem_id_smiles_mapping(version: str) -> Mapping[str, str]:
    """Get a mapping from PubChem compound identifiers to SMILES strings."""
    df = _get_cid_smiles_df(version=version)
    return dict(df.values)


def get_pubchem_smiles_id_mapping(version: str) -> Mapping[str, str]:
    """Get a mapping from SMILES strings to PubChem compound identifiers."""
    df = _get_cid_smiles_df(version=version)
    return {v: k for k, v in df.values}


def get_pubchem_id_to_name(version: str) -> Mapping[str, str]:
    """Get a mapping from PubChem compound identifiers to their titles."""
    # 2 tab-separated columns: compound_id, name
    url = _get_pubchem_extras_url(version, "CID-Title.gz")
    df = ensure_df(PREFIX, url=url, version=version, dtype=str, encoding="latin-1")
    return dict(df.values)


def get_pubchem_id_to_mesh_id(version: str) -> Mapping[str, str]:
    """Get a mapping from PubChem compound identifiers to their equivalent MeSH terms."""
    url = _get_pubchem_extras_url(version, "CID-MeSH")
    df = ensure_df(
        PREFIX,
        url=url,
        version=version,
        dtype=str,
        header=None,
        name="CID-MeSH.tsv",
        names=["pubchem.compound_id", "mesh_id"],
    )
    mesh_name_to_id = get_name_id_mapping("mesh")
    needs_curation = set()
    mesh_ids = []
    for name in df["mesh_id"]:
        mesh_id = mesh_name_to_id.get(name)
        if mesh_id is None:
            if name not in needs_curation:
                needs_curation.add(name)
                logger.debug("[mesh] needs curating: %s", name)
        mesh_ids.append(mesh_id)
    logger.info("[mesh] %d/%d need updating", len(needs_curation), len(mesh_ids))
    df["mesh_id"] = mesh_ids
    return dict(df.values)


def _ensure_cid_name_path(*, version: Optional[str] = None, force: bool = False) -> str:
    if version is None:
        version = bioversions.get_version("pubchem")
    # 2 tab-separated columns: compound_id, name
    cid_name_url = _get_pubchem_extras_url(version, "CID-Title.gz")
    cid_name_path = ensure_path(PREFIX, url=cid_name_url, version=version, force=force)
    return cid_name_path


def get_terms(*, version: str, use_tqdm: bool = True, force: bool = False) -> Iterable[Term]:
    """Get PubChem Compound terms."""
    cid_name_path = _ensure_cid_name_path(version=version, force=force)

    # 2 tab-separated columns: compound_id, synonym
    cid_synonyms_url = _get_pubchem_extras_url(version, "CID-Synonym-filtered.gz")
    cid_synonyms_path = ensure_path(PREFIX, url=cid_synonyms_url, version=version, force=force)

    it = iterate_gzips_together(cid_name_path, cid_synonyms_path)

    if use_tqdm:
        total = 146000000  # got this by reading the exports page
        it = tqdm(it, desc=f"mapping {PREFIX}", unit_scale=True, unit="compound", total=total)
    for identifier, name, raw_synonyms in it:
        reference = Reference(prefix=PREFIX, identifier=identifier, name=name)
        xrefs = []
        synonyms = []
        for synonym in raw_synonyms:
            if synonym.startswith("CHEBI:"):
                xrefs.append(Reference(prefix="chebi", identifier=removeprefix(synonym, "CHEBI:")))
            elif synonym.startswith("CHEMBL"):
                xrefs.append(Reference(prefix="chembl", identifier=synonym))
            elif synonym.startswith("InChI="):
                xrefs.append(Reference(prefix="inchi", identifier=synonym))
            # elif synonym.startswith("SCHEMBL"):
            #     xrefs.append(Reference(prefix="schembl", identifier=synonym))
            else:
                synonyms.append(Synonym(name=synonym))
            # TODO check other xrefs

        term = Term(
            reference=reference,
            synonyms=synonyms,
            xrefs=xrefs,
        )
        yield term


if __name__ == "__main__":
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Configuration for the default priority list."""

import bioregistry

__all__ = [
    "DEFAULT_PRIORITY_LIST",
]

_DEFAULT_PRIORITY_LIST = [
    # Genes
    "ncbigene",
    "hgnc",
    "rgd",
    "mgi",
    "ensembl",
    "uniprot",
    # Chemicals
    # 'inchikey',
    # 'inchi',
    # 'smiles',
    "pubchem.compound",
    "chebi",
    "drugbank",
    "chembl.compound",
    "zinc",
    # protein families and complexes (and famplexes :))
    "complexportal",
    "fplx",
    "ec-code",
    "interpro",
    "pfam",
    "signor",
    # Pathologies/phenotypes
    "mondo",
    "efo",
    "doid",
    "hp",
    # Taxa
    "ncbitaxon",
    # If you can get away from MeSH, do it
    "mesh",
    "icd",
]


def _get_default_priority_list():
    rv = []
    for _entry in _DEFAULT_PRIORITY_LIST:
        _prefix = bioregistry.normalize_prefix(_entry)
        if not _prefix:
            raise RuntimeError(f"unresolved prefix: {_entry}")
        if _prefix in rv:
            raise RuntimeError(f"duplicate found in priority list: {_entry}/{_prefix}")
        rv.append(_prefix)
    return rv


DEFAULT_PRIORITY_LIST = _get_default_priority_list()
del _get_default_priority_list



# -*- coding: utf-8 -*-

"""Constants for PyOBO."""

import logging
import re
from functools import partial
from typing import Callable

import bioversions
import pystow

__all__ = [
    "RAW_DIRECTORY",
    "DATABASE_DIRECTORY",
    "SPECIES_REMAPPING",
    "version_getter",
]

logger = logging.getLogger(__name__)

PYOBO_MODULE = pystow.module("pyobo")
RAW_MODULE = PYOBO_MODULE.module("raw")
RAW_DIRECTORY = RAW_MODULE.base
DATABASE_MODULE = PYOBO_MODULE.module("database")
DATABASE_DIRECTORY = DATABASE_MODULE.base

SPECIES_REMAPPING = {
    "Canis familiaris": "Canis lupus familiaris",
}

GLOBAL_SKIP = {
    "rnao",
    "mo",  # deprecated
    "resid",  # deprecated
    "adw",  # deprecated
}

#: Default prefix
DEFAULT_PREFIX = "debio"
DEFAULT_PATTERN = re.compile("^\\d{7}$")

SOURCE_PREFIX = "source_ns"
SOURCE_ID = "source_id"
RELATION_PREFIX = "relation_ns"
RELATION_ID = "relation_id"
TARGET_PREFIX = "target_ns"
TARGET_ID = "target_id"
PROVENANCE = "source"
RELATION_COLUMNS = [
    SOURCE_PREFIX,
    SOURCE_ID,
    RELATION_PREFIX,
    RELATION_ID,
    TARGET_PREFIX,
    TARGET_ID,
]
XREF_COLUMNS = [SOURCE_PREFIX, SOURCE_ID, TARGET_PREFIX, TARGET_ID, PROVENANCE]

JAVERT_RECORD = "4021477"
JAVERT_FILE = "xrefs.tsv.gz"

OOH_NA_NA_RECORD = "4020486"
OOH_NA_NA_FILE = "names.tsv.gz"

SYNONYMS_RECORD = "4021482"
SYNONYMS_FILE = "synonyms.tsv.gz"

RELATIONS_RECORD = "4625167"
RELATIONS_FILE = "relations.tsv.gz"

PROPERTIES_RECORD = "4625172"
PROPERTIES_FILE = "properties.tsv.gz"

ALTS_DATA_RECORD = "4021476"
ALTS_FILE = "alts.tsv.gz"

DEFINITIONS_RECORD = "4637061"
DEFINITIONS_FILE = "definitions.tsv.gz"

TYPEDEFS_RECORD = "4644013"
TYPEDEFS_FILE = "typedefs.tsv.gz"

SPECIES_RECORD = "5334738"
SPECIES_FILE = "species.tsv.gz"


def version_getter(name: str) -> Callable[[], str]:
    """Make a function appropriate for getting versions."""
    return partial(bioversions.get_version, name)


NCBITAXON_PREFIX = "NCBITaxon"
DATE_FORMAT = "%d:%m:%Y %H:%M"
PROVENANCE_PREFIXES = {
    "pubmed",
    "pmc",
    "doi",
    "biorxiv",
    "chemrxiv",
    "wikipedia",
    "google.patent",
    "agricola",
    "cba",
    "ppr",
    "citexplore",
    "goc",
    "isbn",
    "issn",
}



# -*- coding: utf-8 -*-

"""Converter for ChEMBL.

Run with ``python -m pyobo.sources.chembl -vv``.
"""

import logging
from contextlib import closing
from typing import Iterable

import chembl_downloader

from pyobo.struct import Obo, Reference, Term

__all__ = [
    "ChEMBLCompoundGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "chembl.compound"

QUERY = """\
SELECT
    MOLECULE_DICTIONARY.chembl_id,
    MOLECULE_DICTIONARY.pref_name,
    COMPOUND_STRUCTURES.canonical_smiles,
    COMPOUND_STRUCTURES.standard_inchi,
    COMPOUND_STRUCTURES.standard_inchi_key
FROM MOLECULE_DICTIONARY
    JOIN COMPOUND_STRUCTURES ON MOLECULE_DICTIONARY.molregno == COMPOUND_STRUCTURES.molregno
WHERE
    molecule_dictionary.pref_name IS NOT NULL
"""


# TODO molecule_synonyms table
# TODO molecule_hierarchy table


class ChEMBLCompoundGetter(Obo):
    """An ontology representation of ChEMBL compounds."""

    ontology = "chembl.compound"
    bioversions_key = "chembl"

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Return ChEMBL Compounds as OBO."""
    return ChEMBLCompoundGetter(force=force)


def iter_terms(version: str) -> Iterable[Term]:
    """Iterate over ChEMBL compounds."""
    with chembl_downloader.connect(version=version) as conn:
        logger.info("using connection %s", conn)
        with closing(conn.cursor()) as cursor:
            logger.info("using cursor %s", cursor)
            cursor.execute(QUERY)
            for chembl_id, name, smiles, inchi, inchi_key in cursor.fetchall():
                # TODO add xrefs?
                term = Term.from_triple(prefix=PREFIX, identifier=chembl_id, name=name)
                if smiles:
                    term.append_property("smiles", smiles)
                if inchi:
                    term.append_property("inchi", inchi)
                if inchi_key:
                    term.append_xref(Reference(prefix="inchikey", identifier=inchi_key))
                yield term


if __name__ == "__main__":
    ChEMBLCompoundGetter.cli()



# -*- coding: utf-8 -*-

"""I/O utilities."""

import collections.abc
import csv
import gzip
import logging
import time
from collections import defaultdict
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, Iterable, List, Mapping, Optional, Set, Tuple, TypeVar, Union
from xml.etree.ElementTree import Element

from lxml import etree
from tqdm.auto import tqdm

__all__ = [
    "open_map_tsv",
    "open_multimap_tsv",
    "multidict",
    "multisetdict",
    "write_map_tsv",
    "write_multimap_tsv",
    "write_iterable_tsv",
    "parse_xml_gz",
    "get_writer",
    "open_reader",
    "get_reader",
]

logger = logging.getLogger(__name__)

X = TypeVar("X")
Y = TypeVar("Y")


@contextmanager
def open_reader(path: Union[str, Path], sep: str = "\t"):
    """Open a file and get a reader for it."""
    path = Path(path)
    with gzip.open(path, "rt") if path.suffix == ".gz" else open(path) as file:
        yield get_reader(file, sep=sep)


def get_reader(x, sep: str = "\t"):
    """Get a :func:`csv.reader` with PyOBO default settings."""
    return csv.reader(x, delimiter=sep, quoting=csv.QUOTE_MINIMAL)


def get_writer(x, sep: str = "\t"):
    """Get a :func:`csv.writer` with PyOBO default settings."""
    return csv.writer(x, delimiter=sep, quoting=csv.QUOTE_MINIMAL)


def open_map_tsv(
    path: Union[str, Path], *, use_tqdm: bool = False, has_header: bool = True
) -> Mapping[str, str]:
    """Load a mapping TSV file into a dictionary."""
    with open(path) as file:
        if has_header:
            next(file)  # throw away header
        if use_tqdm:
            file = tqdm(file, desc=f"loading TSV from {path}")
        rv = {}
        for row in get_reader(file):
            if len(row) != 2:
                logger.warning("[%s] malformed row can not be put in dict: %s", path, row)
                continue
            rv[row[0]] = row[1]
        return rv


def open_multimap_tsv(
    path: Union[str, Path],
    *,
    use_tqdm: bool = False,
    has_header: bool = True,
) -> Mapping[str, List[str]]:
    """Load a mapping TSV file that has multiple mappings for each."""
    return multidict(_help_multimap_tsv(path=path, use_tqdm=use_tqdm, has_header=has_header))


def _help_multimap_tsv(
    path: Union[str, Path],
    *,
    use_tqdm: bool = False,
    has_header: bool = True,
) -> Iterable[Tuple[str, str]]:
    with open(path) as file:
        if has_header:
            next(file)  # throw away header
        if use_tqdm:
            file = tqdm(file, desc=f"loading TSV from {path}")
        yield from get_reader(file)


def multidict(pairs: Iterable[Tuple[X, Y]]) -> Mapping[X, List[Y]]:
    """Accumulate a multidict from a list of pairs."""
    rv = defaultdict(list)
    for key, value in pairs:
        rv[key].append(value)
    return dict(rv)


def multisetdict(pairs: Iterable[Tuple[X, Y]]) -> Dict[X, Set[Y]]:
    """Accumulate a multisetdict from a list of pairs."""
    rv = defaultdict(set)
    for key, value in pairs:
        rv[key].add(value)
    return dict(rv)


def write_map_tsv(
    *,
    path: Union[str, Path],
    header: Optional[Iterable[str]] = None,
    rv: Union[Iterable[Tuple[str, str]], Mapping[str, str]],
    sep: str = "\t",
) -> None:
    """Write a mapping dictionary to a TSV file."""
    if isinstance(rv, collections.abc.Mapping):
        write_iterable_tsv(path=path, header=header, it=rv.items(), sep=sep)
    else:
        write_iterable_tsv(path=path, header=header, it=rv, sep=sep)


def write_multimap_tsv(
    *,
    path: Union[str, Path],
    header: Iterable[str],
    rv: Mapping[str, List[str]],
    sep: str = "\t",
) -> None:
    """Write a multiple mapping dictionary to a TSV file."""
    it = ((key, value) for key, values in rv.items() for value in values)
    write_iterable_tsv(path=path, header=header, it=it, sep=sep)


def write_iterable_tsv(
    *,
    path: Union[str, Path],
    header: Optional[Iterable[str]] = None,
    it: Iterable[Tuple[str, ...]],
    sep: str = "\t",
) -> None:
    """Write a mapping dictionary to a TSV file."""
    it = (row for row in it if all(cell is not None for cell in row))
    it = sorted(it)
    with open(path, "w") as file:
        writer = get_writer(file, sep=sep)
        if header is not None:
            writer.writerow(header)
        writer.writerows(it)


def parse_xml_gz(path: Union[str, Path]) -> Element:
    """Parse an XML file from a path to a GZIP file."""
    path = Path(path).resolve()
    t = time.time()
    logger.info("parsing xml from %s", path)
    tree = etree.parse(path.as_posix())  # type:ignore
    logger.info("parsed xml in %.2f seconds", time.time() - t)
    return tree.getroot()



# -*- coding: utf-8 -*-

"""Converter for AGROVOC."""

import pystow
from pystow.utils import read_zipfile_rdf
from rdflib import Graph
from rdflib.namespace import DCTERMS, SKOS

__all__ = [
    "ensure_agrovoc_graph",
]


def ensure_agrovoc_graph(version: str) -> Graph:
    """Download and parse the given version of AGROVOC."""
    url = f"https://agrovoc.fao.org/agrovocReleases/agrovoc_{version}_core.nt.zip"
    path = pystow.ensure("bio", "agrovoc", version, url=url, name="core.nt.zip")
    graph = read_zipfile_rdf(path, inner_path=f"agrovoc_{version}_core.nt", format="nt")
    graph.bind("skosxl", "http://www.w3.org/2008/05/skos-xl#")
    graph.bind("skos", SKOS)
    graph.bind("dcterms", DCTERMS)
    graph.bind("agrovoc", "http://aims.fao.org/aos/agrontology#")
    return graph



# -*- coding: utf-8 -*-

"""High-level API for hierarchies."""

import logging
from functools import lru_cache
from typing import Iterable, Optional, Set, Tuple

import networkx as nx

from .names import get_name
from .properties import get_filtered_properties_mapping
from .relations import get_filtered_relations_df
from ..identifier_utils import wrap_norm_prefix
from ..struct import TypeDef, has_member, is_a, part_of

__all__ = [
    "get_hierarchy",
    "get_subhierarchy",
    "get_descendants",
    "get_ancestors",
    "has_ancestor",
    "is_descendent",
]

from ..struct.reference import Reference

logger = logging.getLogger(__name__)


def get_hierarchy(
    prefix: str,
    *,
    include_part_of: bool = True,
    include_has_member: bool = False,
    extra_relations: Optional[Iterable[TypeDef]] = None,
    properties: Optional[Iterable[str]] = None,
    use_tqdm: bool = False,
    force: bool = False,
    version: Optional[str] = None,
) -> nx.DiGraph:
    """Get hierarchy of parents as a directed graph.

    :param prefix: The name of the namespace.
    :param include_part_of: Add "part of" relations. Only works if the relations are properly
        defined using bfo:0000050 ! part of or bfo:0000051 ! has part
    :param include_has_member: Add "has member" relations. These aren't part of the BFO, but
        are hacked into PyOBO using :data:`pyobo.struct.typedef.has_member` for relationships like
        from protein families to their actual proteins.
    :param extra_relations: Other relations that you want to include in the hierarchy. For
        example, it might be useful to include the positively_regulates
    :param properties: Properties to include in the data part of each node. For example, might want
        to include SMILES strings with the ChEBI tree.
    :param use_tqdm: Show a progress bar
    :param force: should the resources be reloaded when extracting relations?
    :returns: A directional graph representing the hierarchy

    This function thinly wraps :func:`_get_hierarchy_helper` to make it easier to work with the lru_cache mechanism.
    """
    return _get_hierarchy_helper(
        prefix=prefix,
        include_part_of=include_part_of,
        include_has_member=include_has_member,
        extra_relations=tuple(sorted(extra_relations or [], key=lambda t: t.curie)),
        properties=tuple(sorted(properties or [])),
        use_tqdm=use_tqdm,
        force=force,
        version=version,
    )


@lru_cache()
@wrap_norm_prefix
def _get_hierarchy_helper(
    prefix: str,
    *,
    extra_relations: Tuple[TypeDef, ...],
    properties: Tuple[str, ...],
    include_part_of: bool,
    include_has_member: bool,
    use_tqdm: bool,
    force: bool = False,
    version: Optional[str] = None,
) -> nx.DiGraph:
    rv = nx.DiGraph()

    is_a_df = get_filtered_relations_df(
        prefix=prefix,
        relation=is_a,
        use_tqdm=use_tqdm,
        force=force,
        version=version,
    )
    for source_id, target_ns, target_id in is_a_df.values:
        rv.add_edge(f"{prefix}:{source_id}", f"{target_ns}:{target_id}", relation="is_a")

    if include_has_member:
        has_member_df = get_filtered_relations_df(
            prefix=prefix,
            relation=has_member,
            use_tqdm=use_tqdm,
            force=force,
            version=version,
        )
        for target_id, source_ns, source_id in has_member_df.values:
            rv.add_edge(f"{source_ns}:{source_id}", f"{prefix}:{target_id}", relation="is_a")

    if include_part_of:
        part_of_df = get_filtered_relations_df(
            prefix=prefix,
            relation=part_of,
            use_tqdm=use_tqdm,
            force=force,
            version=version,
        )
        for source_id, target_ns, target_id in part_of_df.values:
            rv.add_edge(f"{prefix}:{source_id}", f"{target_ns}:{target_id}", relation="part_of")

        has_part_df = get_filtered_relations_df(
            prefix=prefix,
            relation=part_of,
            use_tqdm=use_tqdm,
            force=force,
            version=version,
        )
        for target_id, source_ns, source_id in has_part_df.values:
            rv.add_edge(f"{source_ns}:{source_id}", f"{prefix}:{target_id}", relation="part_of")

    for relation in extra_relations:
        if not isinstance(relation, (TypeDef, Reference)):
            raise TypeError
        relation_df = get_filtered_relations_df(
            prefix=prefix,
            relation=relation,
            use_tqdm=use_tqdm,
            force=force,
            version=version,
        )
        for source_id, target_ns, target_id in relation_df.values:
            rv.add_edge(
                f"{prefix}:{source_id}", f"{target_ns}:{target_id}", relation=relation.identifier
            )

    for prop in properties:
        props = get_filtered_properties_mapping(
            prefix=prefix, prop=prop, use_tqdm=use_tqdm, force=force
        )
        for identifier, value in props.items():
            curie = f"{prefix}:{identifier}"
            if curie in rv:
                rv.nodes[curie][prop] = value

    return rv


def is_descendent(prefix, identifier, ancestor_prefix, ancestor_identifier) -> bool:
    """Check that the first identifier has the second as a descendent.

    Check that go:0070246 ! natural killer cell apoptotic process is a
    descendant of go:0006915 ! apoptotic process::
    >>> assert is_descendent('go', '0070246', 'go', '0006915')
    """
    descendants = get_descendants(ancestor_prefix, ancestor_identifier)
    return descendants is not None and f"{prefix}:{identifier}" in descendants


@lru_cache()
def get_descendants(
    prefix: str,
    identifier: str,
    include_part_of: bool = True,
    include_has_member: bool = False,
    use_tqdm: bool = False,
    force: bool = False,
    **kwargs,
) -> Optional[Set[str]]:
    """Get all of the descendants (children) of the term as CURIEs."""
    hierarchy = get_hierarchy(
        prefix=prefix,
        include_has_member=include_has_member,
        include_part_of=include_part_of,
        use_tqdm=use_tqdm,
        force=force,
        **kwargs,
    )
    curie = f"{prefix}:{identifier}"
    if curie not in hierarchy:
        return None
    return nx.ancestors(hierarchy, curie)  # note this is backwards


def has_ancestor(prefix, identifier, ancestor_prefix, ancestor_identifier) -> bool:
    """Check that the first identifier has the second as an ancestor.

    Check that go:0008219 ! cell death is an ancestor of go:0006915 ! apoptotic process::
    >>> assert has_ancestor('go', '0006915', 'go', '0008219')
    """
    ancestors = get_ancestors(prefix, identifier)
    return ancestors is not None and f"{ancestor_prefix}:{ancestor_identifier}" in ancestors


@lru_cache()
def get_ancestors(
    prefix: str,
    identifier: str,
    include_part_of: bool = True,
    include_has_member: bool = False,
    use_tqdm: bool = False,
    force: bool = False,
    **kwargs,
) -> Optional[Set[str]]:
    """Get all of the ancestors (parents) of the term as CURIEs."""
    hierarchy = get_hierarchy(
        prefix=prefix,
        include_has_member=include_has_member,
        include_part_of=include_part_of,
        use_tqdm=use_tqdm,
        force=force,
        **kwargs,
    )
    curie = f"{prefix}:{identifier}"
    if curie not in hierarchy:
        return None
    return nx.descendants(hierarchy, curie)  # note this is backwards


def get_subhierarchy(
    prefix: str,
    identifier: str,
    include_part_of: bool = True,
    include_has_member: bool = False,
    use_tqdm: bool = False,
    force: bool = False,
    **kwargs,
) -> nx.DiGraph:
    """Get the subhierarchy for a given node."""
    hierarchy = get_hierarchy(
        prefix=prefix,
        include_has_member=include_has_member,
        include_part_of=include_part_of,
        use_tqdm=use_tqdm,
        force=force,
        **kwargs,
    )
    logger.info(
        "getting descendants of %s:%s ! %s", prefix, identifier, get_name(prefix, identifier)
    )
    curies = nx.ancestors(hierarchy, f"{prefix}:{identifier}")  # note this is backwards
    logger.info("inducing subgraph")
    sg = hierarchy.subgraph(curies).copy()
    logger.info("subgraph has %d nodes/%d edges", sg.number_of_nodes(), sg.number_of_edges())
    return sg



# -*- coding: utf-8 -*-

"""Extract and convert BioGRID identifiers."""

from typing import Mapping, Optional

import bioversions
import pandas as pd

from pyobo.constants import version_getter
from pyobo.resources.ncbitaxon import get_ncbitaxon_id
from pyobo.utils.cache import cached_mapping
from pyobo.utils.path import ensure_df, prefix_directory_join

PREFIX = "biogrid"
BASE_URL = "https://downloads.thebiogrid.org/Download/BioGRID/Release-Archive"

taxonomy_remapping = {  # so much for official names
    "Canis familiaris": "9615",  # Canis lupus familiaris
    "Human Herpesvirus 1": "10298",  # Human alphaherpesvirus 1
    "Human Herpesvirus 3": "10335",  # Human alphaherpesvirus 3
    "Murid Herpesvirus 1": "10366",  # Murid betaherpesvirus 1
    "Human Herpesvirus 4": "10376",  # Human gammaherpesvirus 4
    "Hepatitus C Virus": "11103",  # Hepacivirus C
    "Human Immunodeficiency Virus 1": "11676",  # Human immunodeficiency virus 1
    "Human Immunodeficiency Virus 2": "11709",  # Human immunodeficiency virus 2
    "Human Herpesvirus 2": "10310",  # Human alphaherpesvirus 2
    "Human Herpesvirus 5": "10359",  # Human betaherpesvirus 5
    "Human Herpesvirus 6A": "32603",  # Human betaherpesvirus 6A
    "Human Herpesvirus 6B": "32604",  # Human betaherpesvirus 6B
    "Human Herpesvirus 7": "10372",  # Human betaherpesvirus 7
    "Human Herpesvirus 8": "37296",  # Human gammaherpesvirus 8
    "Emericella nidulans": "162425",  # Aspergillus nidulans
    "Bassica campestris": "145471",  # Brassica rapa subsp. oleifera (was a typo)
    "Tarsius syrichta": "1868482",  # Carlito syrichta
    "Felis Catus": "9685",  # Felis catus
    "Vaccinia Virus": "10245",  # Vaccinia virus
    "Simian Virus 40": "1891767",  # Macaca mulatta polyomavirus 1
    "Simian Immunodeficiency Virus": "11723",  # Simian immunodeficiency virus
    "Tobacco Mosaic Virus": "12242",  # Tobacco mosaic virus
    # Not in my current dump, but definitely there!
    "Severe acute respiratory syndrome coronavirus 2": "2697049",  # Severe acute respiratory syndrome coronavirus 2
    "Middle-East Respiratory Syndrome-related Coronavirus": "1335626",
}


def _lookup(name: str) -> Optional[str]:
    if name in taxonomy_remapping:
        return taxonomy_remapping[name]
    return get_ncbitaxon_id(name)


def get_df() -> pd.DataFrame:
    """Get the BioGRID identifiers mapping dataframe."""
    version = bioversions.get_version("biogrid")
    url = f"{BASE_URL}/BIOGRID-{version}/BIOGRID-IDENTIFIERS-{version}.tab.zip"
    df = ensure_df(PREFIX, url=url, skiprows=28, dtype=str, version=version)
    df["taxonomy_id"] = df["ORGANISM_OFFICIAL_NAME"].map(_lookup)
    return df


@cached_mapping(
    path=prefix_directory_join(
        PREFIX, "cache", "xrefs", name="ncbigene.tsv", version=version_getter(PREFIX)
    ),
    header=["biogrid_id", "ncbigene_id"],
)
def get_ncbigene_mapping() -> Mapping[str, str]:
    """Get BioGRID to NCBIGene mapping.

    Is basically equivalent to:

    .. code-block:: python

        from pyobo import get_filtered_xrefs
        biogrid_ncbigene_mapping = get_filtered_xrefs('biogrid', 'ncbigene')
    """
    df = get_df()
    df = df.loc[df["IDENTIFIER_TYPE"] == "ENTREZ_GENE", ["BIOGRID_ID", "IDENTIFIER_VALUE"]]
    return dict(df.values)



# -*- coding: utf-8 -*-

"""Swisslipids."""

from typing import Iterable

import pandas as pd
from tqdm.auto import tqdm

from pyobo import Obo, SynonymTypeDef, Term
from pyobo.utils.path import ensure_df

__all__ = [
    "SLMGetter",
]

PREFIX = "slm"
COLUMNS = [
    "Lipid ID",
    "Level",
    "Name",
    "Abbreviation*",
    "Synonyms*",
    "Lipid class*",
    "Parent",
    "Components*",
    "SMILES (pH7.3)",
    "InChI (pH7.3)",
    "InChI key (pH7.3)",
    # "Formula (pH7.3)", "Charge (pH7.3)", "Mass (pH7.3)",
    # "Exact Mass (neutral form)", "Exact m/z of [M.]+", "Exact m/z of [M+H]+", "Exact m/z of [M+K]+ ",
    # "Exact m/z of [M+Na]+", "Exact m/z of [M+Li]+", "Exact m/z of [M+NH4]+", "Exact m/z of [M-H]-",
    # "Exact m/z of [M+Cl]-", "Exact m/z of [M+OAc]- ",
    "CHEBI",
    "LIPID MAPS",
    "HMDB",
    "PMID",
]

abreviation_type = SynonymTypeDef.from_text("abbreviation")


class SLMGetter(Obo):
    """An ontology representation of SwissLipid's lipid nomenclature."""

    ontology = bioversions_key = PREFIX
    synonym_typedefs = [abreviation_type]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(force=force, version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Get SwissLipids as OBO."""
    return SLMGetter(force=force)


def iter_terms(version: str, force: bool = False):
    """Iterate over SwissLipids terms."""
    df = ensure_df(
        prefix=PREFIX,
        url="https://www.swisslipids.org/api/file.php?cas=download_files&file=lipids.tsv",
        version=version,
        name="lipids.tsv.gz",
        encoding="cp1252",
        force=force,
    )
    for (
        identifier,
        level,
        name,
        abbreviation,
        synonyms,
        _cls,
        _parent,
        _components,
        smiles,
        inchi,
        inchikey,
        chebi_id,
        lipidmaps_id,
        hmdb_id,
        pmids,
    ) in tqdm(
        df[COLUMNS].values, desc=f"[{PREFIX}] generating terms", unit_scale=True, unit="lipid"
    ):
        if identifier.startswith("SLM:"):
            identifier = identifier[len("SLM:") :]
        else:
            raise ValueError(identifier)
        term = Term.from_triple(PREFIX, identifier, name)
        term.append_property("level", level)
        if pd.notna(abbreviation):
            term.append_synonym(abbreviation, type=abreviation_type)
        if pd.notna(synonyms):
            for synonym in synonyms.split("|"):
                term.append_synonym(synonym.strip())
        if pd.notna(smiles):
            term.append_property("smiles", smiles)
        if pd.notna(inchi) and inchi != "InChI=none":
            if inchi.startswith("InChI="):
                inchi = inchi[len("InChI=") :]
            term.append_property("inchi", inchi)
        if pd.notna(inchikey):
            if inchikey.startswith("InChIKey="):
                inchikey = inchikey[len("InChIKey=") :]
            term.append_property("inchikey", inchikey)
        if pd.notna(chebi_id):
            term.append_xref(("chebi", chebi_id))
        if pd.notna(lipidmaps_id):
            term.append_xref(("lipidmaps", lipidmaps_id))
        if pd.notna(hmdb_id):
            term.append_xref(("hmdb", hmdb_id))
        if pd.notna(pmids):
            for pmid in pmids.split("|"):
                term.append_provenance(("pubmed", pmid))
        # TODO how to handle class, parents, and components?
        yield term


if __name__ == "__main__":
    get_obo().write_default(write_obo=True, use_tqdm=True)



# -*- coding: utf-8 -*-

"""Resource utilities for PyOBO."""

from functools import lru_cache
from typing import Sequence

import click
import pandas as pd
from more_click import verbose_option
from zenodo_client import Zenodo

from .constants import (
    ALTS_DATA_RECORD,
    ALTS_FILE,
    DEFINITIONS_FILE,
    DEFINITIONS_RECORD,
    JAVERT_FILE,
    JAVERT_RECORD,
    OOH_NA_NA_FILE,
    OOH_NA_NA_RECORD,
    PROPERTIES_FILE,
    PROPERTIES_RECORD,
    RELATIONS_FILE,
    RELATIONS_RECORD,
    SPECIES_FILE,
    SPECIES_RECORD,
    SYNONYMS_FILE,
    SYNONYMS_RECORD,
)

__all__ = [
    "ensure_inspector_javert",
    "ensure_inspector_javert_df",
    "ensure_ooh_na_na",
    "ensure_alts",
    "ensure_synonyms",
    "ensure_definitions",
    "ensure_properties",
    "ensure_relations",
    "ensure_species",
]


@lru_cache(maxsize=1)
def _get_zenodo() -> Zenodo:
    """Get the cached Zenodo client."""
    return Zenodo()


def _get_parts(_concept_rec_id, _record_id, version) -> Sequence[str]:
    """Get sequence to use in :func:`pystow.ensure`.

    .. note:: Corresponds to :data:`pyobo.constants.DATABASE_MODULE`.
    """
    return ["pyobo", "database", version]


def _ensure(record_id, path, force: bool = False) -> str:
    rv = _get_zenodo().download_latest(
        record_id=record_id, path=path, parts=_get_parts, force=force
    )
    return rv.as_posix()


def ensure_ooh_na_na(force: bool = False) -> str:
    """Ensure that the Ooh Na Na Nomenclature Database is downloaded/built.

    .. seealso:: :data:`pyobo.constants.OOH_NA_NA_RECORD`
    """
    return _ensure(record_id=OOH_NA_NA_RECORD, path=OOH_NA_NA_FILE, force=force)


def ensure_inspector_javert(force: bool = False) -> str:
    """Ensure that the Inspector Javert's Xref Database is downloaded/built.

    .. seealso:: :data:`pyobo.constants.JAVERT_RECORD`
    """
    return _ensure(record_id=JAVERT_RECORD, path=JAVERT_FILE, force=force)


def ensure_inspector_javert_df(force: bool = False) -> pd.DataFrame:
    """Ensure the inspector javert database as a dataframe."""
    path = ensure_inspector_javert(force=force)
    return pd.read_csv(path, sep="\t", dtype=str)


def ensure_synonyms(force: bool = False) -> str:
    """Ensure that the Synonym Database is downloaded/built.

    .. seealso:: :data:`pyobo.constants.SYNONYMS_RECORD`
    """
    return _ensure(record_id=SYNONYMS_RECORD, path=SYNONYMS_FILE, force=force)


def ensure_alts(force: bool = False) -> str:
    """Ensure that the alt data is downloaded/built.

    .. seealso:: :data:`pyobo.constants.ALTS_DATA_RECORD`
    """
    return _ensure(record_id=ALTS_DATA_RECORD, path=ALTS_FILE, force=force)


def ensure_species(force: bool = False) -> str:
    """Ensure that the species data is downloaded/built.

    .. seealso:: :data:`pyobo.constants.SPECIES_RECORD`
    """
    return _ensure(record_id=SPECIES_RECORD, path=SPECIES_FILE, force=force)


def ensure_definitions(force: bool = False) -> str:
    """Ensure that the definitions data is downloaded/built.

    .. seealso:: :data:`pyobo.constants.DEFINITIONS_RECORD`
    """
    return _ensure(record_id=DEFINITIONS_RECORD, path=DEFINITIONS_FILE, force=force)


def ensure_properties(force: bool = False) -> str:
    """Ensure that the properties data is downloaded/built.

    .. seealso:: :data:`pyobo.constants.PROPERTIES_RECORD`
    """
    return _ensure(record_id=PROPERTIES_RECORD, path=PROPERTIES_FILE, force=force)


def ensure_relations(force: bool = False) -> str:
    """Ensure that the relations data is downloaded/built.

    .. seealso:: :data:`pyobo.constants.RELATIONS_RECORD`
    """
    return _ensure(record_id=RELATIONS_RECORD, path=RELATIONS_FILE, force=force)


@click.command()
@verbose_option
@click.option("-f", "--force", is_flag=True)
def main(force: bool):
    """Ensure resources are available."""
    for f in [
        ensure_ooh_na_na,
        ensure_synonyms,
        ensure_alts,
        ensure_inspector_javert,
        ensure_definitions,
        ensure_properties,
        ensure_relations,
    ]:
        doc = f.__doc__
        if doc is None:
            continue
        click.secho(doc.splitlines()[0], fg="green", bold=True)
        path = f(force=force)
        click.echo(path)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Cross references from cbms2019.

.. seealso:: https://github.com/pantapps/cbms2019
"""

import pandas as pd

from pyobo.constants import (
    PROVENANCE,
    SOURCE_ID,
    SOURCE_PREFIX,
    TARGET_ID,
    TARGET_PREFIX,
    XREF_COLUMNS,
)

__all__ = [
    "get_cbms2019_xrefs_df",
]

#: Columns: DOID, DO name, xref xb, xref ix
base_url = "https://raw.githubusercontent.com/pantapps/cbms2019/master"
doid_to_all = f"{base_url}/mesh_icd10cm_via_do_not_mapped_umls.tsv"
#: Columns: SNOMEDCT_ID, SNOMEDCIT_NAME, ICD10CM_ID, ICD10CM_NAME, MESH_ID
all_to_all = f"{base_url}/mesh_icd10cm_via_snomedct_not_mapped_umls.tsv"
#: Columns: DOID, DO name, xref xb, xref ix
doid_to_all_2 = f"{base_url}/mesh_snomedct_via_do_not_mapped_umls.tsv"
#: Columns: SNOMEDCT_ID, SNOMEDCIT_NAME, ICD10CM_ID, ICD10CM_NAME, MESH_ID
all_to_all_2 = f"{base_url}/mesh_snomedct_via_icd10cm_not_mapped_umls.tsv"

NSM = {
    "MESH": "mesh",
    "ICD10CM": "icd",
    "SNOMEDCT_US_2016_03_01": "snomedct",
}


def _get_doid(url: str) -> pd.DataFrame:
    df = pd.read_csv(url, sep="\t", usecols=["DO_ID", "resource", "resource_ID"])
    df.columns = [SOURCE_ID, TARGET_PREFIX, TARGET_ID]

    df[SOURCE_PREFIX] = "doid"
    df[SOURCE_ID] = df[SOURCE_ID].map(lambda s: s[len("DOID:") :])
    df[PROVENANCE] = url
    df[TARGET_PREFIX] = df[TARGET_PREFIX].map(NSM.get)
    df = df[XREF_COLUMNS]
    return df


def _get_mesh_to_icd_via_doid() -> pd.DataFrame:
    return _get_doid(doid_to_all)


def _get_mesh_to_icd_via_snomedct() -> pd.DataFrame:
    df = pd.read_csv(all_to_all, sep="\t", usecols=["SNOMEDCT_ID", "ICD10CM_ID", "MESH_ID"])
    rows = []
    for snomedct_id, icd_id, mesh_id in df.values:
        rows.append(("mesh", mesh_id, "snomedct", snomedct_id, all_to_all))
        rows.append(("snomedct", snomedct_id, "icd", icd_id, all_to_all))
    return pd.DataFrame(rows, columns=XREF_COLUMNS)


def _get_mesh_to_snomedct_via_doid() -> pd.DataFrame:
    return _get_doid(doid_to_all_2)


def _get_mesh_to_snomedct_via_icd() -> pd.DataFrame:
    df = pd.read_csv(
        all_to_all_2,
        sep="\t",
        usecols=["SNOMEDCT_ID", "ICD10CM_ID", "MESH_ID"],
        dtype={"SNOMEDCT_ID": float},
    )
    rows = []
    for snomedct_id, icd_id, mesh_id in df.values:
        snomedct_id = str(int(snomedct_id))
        rows.append(("mesh", mesh_id, "icd", icd_id, all_to_all))
        rows.append(("icd", icd_id, "snomedct", snomedct_id, all_to_all))
    return pd.DataFrame(rows, columns=XREF_COLUMNS)


def get_cbms2019_xrefs_df() -> pd.DataFrame:
    """Get all CBMS2019 xrefs."""
    return pd.concat(
        [
            _get_mesh_to_icd_via_doid(),
            _get_mesh_to_icd_via_snomedct(),
            _get_mesh_to_snomedct_via_doid(),
            _get_mesh_to_snomedct_via_icd(),
        ]
    ).drop_duplicates()



# -*- coding: utf-8 -*-

"""Get xrefs from PubChem Compound to MeSH."""

from typing import Optional

import bioversions
import pandas as pd

from ...constants import XREF_COLUMNS
from ...sources.pubchem import _get_pubchem_extras_url, get_pubchem_id_to_mesh_id

__all__ = [
    "get_pubchem_mesh_df",
]


def get_pubchem_mesh_df(version: Optional[str] = None) -> pd.DataFrame:
    """Get PubChem Compound-MeSH xrefs."""
    if version is None:
        version = bioversions.get_version("pubchem")
    cid_mesh_url = _get_pubchem_extras_url(version, "CID-MeSH")
    return pd.DataFrame(
        [
            ("pubchem.compound", k, "mesh", v, cid_mesh_url)
            for k, v in get_pubchem_id_to_mesh_id(version=version).items()
        ],
        columns=XREF_COLUMNS,
    )



# -*- coding: utf-8 -*-

"""High-level API for synonyms."""

import logging
from functools import lru_cache
from typing import Mapping, Optional

import bioregistry
import pandas as pd
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from .utils import get_version
from ..constants import TARGET_ID, TARGET_PREFIX
from ..getters import get_ontology
from ..identifier_utils import wrap_norm_prefix
from ..utils.cache import cached_df, cached_mapping
from ..utils.path import prefix_cache_join

__all__ = [
    "get_xrefs_df",
    "get_filtered_xrefs",
    "get_xref",
    "get_xrefs",
    "get_sssom_df",
]

logger = logging.getLogger(__name__)


@wrap_norm_prefix
def get_xref(prefix: str, identifier: str, new_prefix: str, flip: bool = False) -> Optional[str]:
    """Get the xref with the new prefix if a direct path exists."""
    filtered_xrefs = get_filtered_xrefs(prefix, new_prefix, flip=flip)
    return filtered_xrefs.get(identifier)


@lru_cache()
@wrap_norm_prefix
def get_filtered_xrefs(
    prefix: str,
    xref_prefix: str,
    flip: bool = False,
    *,
    use_tqdm: bool = False,
    force: bool = False,
    strict: bool = False,
    version: Optional[str] = None,
) -> Mapping[str, str]:
    """Get xrefs to a given target."""
    if version is None:
        version = get_version(prefix)
    path = prefix_cache_join(prefix, "xrefs", name=f"{xref_prefix}.tsv", version=version)
    all_xrefs_path = prefix_cache_join(prefix, name="xrefs.tsv", version=version)
    header = [f"{prefix}_id", f"{xref_prefix}_id"]

    @cached_mapping(path=path, header=header, use_tqdm=use_tqdm, force=force)
    def _get_mapping() -> Mapping[str, str]:
        if all_xrefs_path.is_file():
            logger.info("[%s] loading pre-cached xrefs", prefix)
            df = pd.read_csv(all_xrefs_path, sep="\t", dtype=str)
            logger.info("[%s] filtering pre-cached xrefs", prefix)
            df = df.loc[df[TARGET_PREFIX] == xref_prefix, [f"{prefix}_id", TARGET_ID]]
            return dict(df.values)

        logger.info("[%s] no cached xrefs found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, strict=strict, version=version)
        return ontology.get_filtered_xrefs_mapping(xref_prefix, use_tqdm=use_tqdm)

    rv = _get_mapping()
    if flip:
        return {v: k for k, v in rv.items()}
    return rv


get_xrefs = get_filtered_xrefs


@wrap_norm_prefix
def get_xrefs_df(
    prefix: str,
    *,
    use_tqdm: bool = False,
    force: bool = False,
    strict: bool = False,
    version: Optional[str] = None,
) -> pd.DataFrame:
    """Get all xrefs."""
    if version is None:
        version = get_version(prefix)
    path = prefix_cache_join(prefix, name="xrefs.tsv", version=version)

    @cached_df(path=path, dtype=str, force=force)
    def _df_getter() -> pd.DataFrame:
        logger.info("[%s] no cached xrefs found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, strict=strict, version=version)
        return ontology.get_xrefs_df(use_tqdm=use_tqdm)

    return _df_getter()


@wrap_norm_prefix
def get_sssom_df(
    prefix: str,
    *,
    predicate_id: str = "oboinowl:hasDbXref",
    justification: str = "sempav:UnspecifiedMatching",
    **kwargs,
) -> pd.DataFrame:
    r"""Get xrefs from a source as an SSSOM dataframe.

    :param prefix: The ontology to look in for xrefs
    :param predicate_id: The predicate used in the SSSOM document. By default, ontologies
        don't typically ascribe semantics to xrefs so ``oboinowl:hasDbXref`` is used
    :param justification: The justification for the mapping. By default, ontologies
        don't typically ascribe semantics, so this is left with `sempav:UnspecifiedMatching`
    :returns: A SSSOM-compliant dataframe of xrefs

    For example, if you want to get UMLS as an SSSOM dataframe, you can do

    >>> import pyobo
    >>> df = pyobo.get_sssom_df("umls")
    >>> df.to_csv("umls.sssom.tsv", sep="\t", index=False)

    .. note:: This assumes the Bioregistry as the prefix map
    """
    from .names import get_name

    df = get_xrefs_df(prefix=prefix, **kwargs)
    with logging_redirect_tqdm():
        rows = [
            (
                bioregistry.curie_to_str(prefix, source_id),
                get_name(prefix, source_id) or "",
                bioregistry.curie_to_str(target_prefix, target_id),
                get_name(target_prefix, target_id),
                predicate_id,
                justification,
            )
            for source_id, target_prefix, target_id in tqdm(
                df.values, unit="mapping", unit_scale=True
            )
        ]
    return pd.DataFrame(
        rows,
        columns=[
            "subject_id",
            "subject_label",
            "object_id",
            "object_label",
            "predicate_id",
            "mapping_justification",
        ],
    )



# -*- coding: utf-8 -*-

"""CLI for PyOBO apps."""

import click

from .gilda.cli import main as gilda_main
from .mapper.cli import main as mapper_main

__all__ = [
    "main",
]


@click.group(name="apps")
def main():
    """Apps."""


main.add_command(gilda_main)
main.add_command(mapper_main)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Utilities for the CLI."""

import datetime
import pathlib

import click
import pandas as pd

from ..constants import DATABASE_DIRECTORY

__all__ = [
    "echo_df",
    "directory_option",
    "zenodo_option",
    "force_option",
    "prefix_argument",
    "no_strict_option",
]


def echo_df(df: pd.DataFrame) -> None:
    """Echo a dataframe via the pager."""
    click.echo_via_pager("\n".join("\t".join(row) for row in df.values))


def get_default_directory() -> pathlib.Path:
    """Get the default database build directory."""
    rv = DATABASE_DIRECTORY / datetime.datetime.today().strftime("%Y-%m-%d")
    rv.mkdir(exist_ok=True, parents=True)
    return rv


directory_option = click.option(
    "--directory",
    type=click.Path(dir_okay=True, file_okay=False, exists=True),
    default=get_default_directory,
    help=f"Build location. Defaults to {DATABASE_DIRECTORY}/<today>",
)
zenodo_option = click.option("--zenodo", is_flag=True, help="Automatically upload to zenodo")
no_strict_option = click.option(
    "-x", "--no-strict", is_flag=True, help="Turn off failure on bad CURIEs"
)
prefix_argument = click.argument("prefix")
force_option = click.option("-f", "--force", is_flag=True)
version_option = click.option(
    "--version",
    help="Explicit version of the data. If not given, the most recent will be looked up.",
)



# -*- coding: utf-8 -*-

"""Pipeline for building a large ontology graph."""

import logging

import bioregistry
import networkx as nx
from tqdm.auto import tqdm

from pyobo import get_hierarchy
from pyobo.getters import SKIP
from pyobo.resource_utils import ensure_inspector_javert_df

logger = logging.getLogger(__name__)


def bens_magical_ontology(use_tqdm: bool = True) -> nx.DiGraph:
    """Make a super graph containing is_a, part_of, and xref relationships."""
    rv = nx.DiGraph()

    df = ensure_inspector_javert_df()
    for source_ns, source_id, target_ns, target_id, provenance in df.values:
        rv.add_edge(
            f"{source_ns}:{source_id}",
            f"{target_ns}:{target_id}",
            relation="xref",
            provenance=provenance,
        )

    logger.info("getting hierarchies")
    it = tqdm(sorted(bioregistry.read_registry()), desc="Entries", disable=not use_tqdm)
    for prefix in it:
        if bioregistry.is_deprecated(prefix) or prefix in SKIP:
            continue
        if use_tqdm:
            it.set_postfix({"prefix": prefix})

        hierarchy = get_hierarchy(prefix, include_has_member=True, include_part_of=True)
        rv.add_edges_from(hierarchy.edges(data=True))

    # TODO include translates_to, transcribes_to, and has_variant

    return rv



# -*- coding: utf-8 -*-

"""Convert PFAM to OBO."""

from typing import Iterable

import pandas as pd

from ..struct import Obo, Reference, Term
from ..utils.path import ensure_df

__all__ = [
    "PfamGetter",
]

PREFIX = "pfam"

CLAN_MAPPING_HEADER = [
    "family_id",
    "clan_id",
    "clan_name",
    "family_name",
    "family_summary",
]


def get_pfam_clan_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get PFAM + clans."""
    url = f"ftp://ftp.ebi.ac.uk/pub/databases/Pfam/releases/Pfam{version}/Pfam-A.clans.tsv.gz"
    return ensure_df(
        PREFIX,
        url=url,
        compression="gzip",
        names=CLAN_MAPPING_HEADER,
        version=version,
        dtype=str,
        force=force,
    )


class PfamGetter(Obo):
    """An ontology representation of Pfam's protein family nomenclature."""

    ontology = bioversions_key = PREFIX

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get PFAM as OBO."""
    return PfamGetter(force=force)


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Iterate PFAM terms."""
    df = get_pfam_clan_df(version=version, force=force)
    for family_identifier, clan_id, clan_name, family_name, definition in df.values:
        parents = []
        if pd.notna(clan_id) and pd.notna(clan_name):
            parents.append(Reference(prefix="pfam.clan", identifier=clan_id, name=clan_name))
        yield Term(
            reference=Reference(prefix=PREFIX, identifier=family_identifier, name=family_name),
            definition=definition,
            parents=parents,
        )


if __name__ == "__main__":
    get_obo().write_default()



# -*- coding: utf-8 -*-

"""Miscellaneous utilities."""

import gzip
import logging
import os
from datetime import datetime
from subprocess import check_output  # noqa:S404
from typing import Optional

__all__ = [
    "obo_to_obograph",
    "obo_to_owl",
    "cleanup_version",
]


logger = logging.getLogger(__name__)


def obo_to_obograph(obo_path, obograph_path) -> None:
    """Convert an OBO file to OBO Graph file with pronto."""
    import pronto

    ontology = pronto.Ontology(obo_path)
    with gzip.open(obograph_path, "wb") as file:
        ontology.dump(file, format="json")


def obo_to_owl(obo_path, owl_path, owl_format: str = "ofn"):
    """Convert an OBO file to an OWL file with ROBOT."""
    args = ["robot", "convert", "-i", obo_path, "-o", owl_path, "--format", owl_format]
    ret = check_output(  # noqa:S603
        args,
        cwd=os.path.dirname(__file__),
    )
    return ret.decode()


BIZARRE_LOGGED = set()


def cleanup_version(data_version: str, prefix: str) -> Optional[str]:
    """Clean the version information."""
    if data_version.endswith(".owl"):
        data_version = data_version[: -len(".owl")]
    if data_version.endswith(prefix):
        data_version = data_version[: -len(prefix)]
    if data_version.startswith("releases/"):
        data_version = data_version[len("releases/") :]
    if prefix == "orth":
        # TODO add bioversions for this
        return "2"
    if data_version.startswith("http://www.orpha.net/version"):
        return data_version[len("http://www.orpha.net/version") :]
    if data_version.startswith("http://humanbehaviourchange.org/ontology/bcio.owl/"):
        return data_version[len("http://humanbehaviourchange.org/ontology/bcio.owl/") :]
    if data_version.startswith("http://www.ebi.ac.uk/efo/releases/v"):
        return data_version[len("http://www.ebi.ac.uk/efo/releases/v") :].split("/")[0]
    if data_version.startswith("http://www.ebi.ac.uk/swo/swo.owl/"):
        return data_version[len("http://www.ebi.ac.uk/swo/swo.owl/") :].split("/")[0]
    if data_version.startswith("http://purl.org/pav/"):
        return data_version[len("http://purl.org/pav/") :]
    if data_version.startswith("http://semanticscience.org/ontology/sio/v"):
        return data_version[len("http://semanticscience.org/ontology/sio/v")].split("/")[0]
    if data_version.startswith("http://ontology.neuinfo.org/NIF/ttl/nif/version/"):
        return data_version[len("http://ontology.neuinfo.org/NIF/ttl/nif/version/") :].split("/")[0]
    if data_version.startswith("http://identifiers.org/combine.specifications/teddy.rel-"):
        return data_version[len("http://identifiers.org/combine.specifications/teddy.rel-") :]
    if data_version.replace(".", "").isnumeric():
        return data_version  # consecutive, major.minor, or semantic versioning
    for v in reversed(data_version.split("/")):
        v = v.strip()
        try:
            datetime.strptime(v, "%Y-%m-%d")
        except ValueError:
            continue
        else:
            return v
    if (prefix, data_version) not in BIZARRE_LOGGED:
        logger.debug("[%s] bizarre version: %s", prefix, data_version)
        BIZARRE_LOGGED.add((prefix, data_version))
    return data_version



# -*- coding: utf-8 -*-

"""Converter for RGD."""

import logging
from typing import Iterable, Optional

import pandas as pd
from tqdm.auto import tqdm

from pyobo.struct import (
    Obo,
    Reference,
    Synonym,
    SynonymTypeDef,
    Term,
    from_species,
    has_gene_product,
    transcribes_to,
)
from pyobo.utils.path import ensure_df

logger = logging.getLogger(__name__)
PREFIX = "rgd"

old_symbol_type = SynonymTypeDef.from_text("old_symbol")
old_name_type = SynonymTypeDef.from_text("old_name")

# NOTE unigene id was discontinue in January 18th, 2021 dump

GENES_URL = "https://download.rgd.mcw.edu/data_release/GENES.RAT.txt"
GENES_HEADER = [
    "GENE_RGD_ID",
    "SYMBOL",
    "NAME",
    "GENE_DESC",
    "CHROMOSOME_CELERA",
    "CHROMOSOME_[oldAssembly#] chromosome for the old reference assembly",
    "CHROMOSOME_[newAssembly#] chromosome for the current reference assembly",
    "FISH_BAND",
    "START_POS_CELERA",
    "STOP_POS_CELERA",
    "STRAND_CELERA",
    "START_POS_[oldAssembly#]",
    "STOP_POS_[oldAssembly#]",
    "STRAND_[oldAssembly#]",
    "START_POS_[newAssembly#]",
    "STOP_POS_[newAssembly#]",
    "STRAND_[newAssembly#]",
    "CURATED_REF_RGD_ID",
    "CURATED_REF_PUBMED_ID",
    "UNCURATED_PUBMED_ID",
    "NCBI_GENE_ID",
    "UNIPROT_ID",
    "UNCURATED_REF_MEDLINE_ID",
    "GENBANK_NUCLEOTIDE",
    "TIGR_ID",
    "GENBANK_PROTEIN",
    "SSLP_RGD_ID",
    "SSLP_SYMBOL",
    "OLD_SYMBOL",
    "OLD_NAME",
    "QTL_RGD_ID",
    "QTL_SYMBOL",
    "NOMENCLATURE_STATUS",
    "SPLICE_RGD_ID",
    "SPLICE_SYMBOL",
    "GENE_TYPE",
    "ENSEMBL_ID",
]


class RGDGetter(Obo):
    """An ontology representation of RGD's rat gene nomenclature."""

    bioversions_key = ontology = PREFIX
    typedefs = [from_species, transcribes_to, has_gene_product]
    synonym_typedefs = [old_name_type, old_symbol_type]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return get_terms(force=force, version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Get RGD as OBO."""
    return RGDGetter(force=force)


namespace_to_column = [
    ("ensembl", "ENSEMBL_ID"),
    ("uniprot", "UNIPROT_ID"),
    ("ncbigene", "NCBI_GENE_ID"),
]


def get_terms(force: bool = False, version: Optional[str] = None) -> Iterable[Term]:
    """Get RGD terms."""
    df = ensure_df(
        PREFIX,
        url=GENES_URL,
        sep="\t",
        header=0,
        comment="#",
        dtype=str,
        force=force,
        version=version,
        quoting=3,
        error_bad_lines=False,
    )
    for _, row in tqdm(
        df.iterrows(), total=len(df.index), desc=f"Mapping {PREFIX}", unit_scale=True
    ):
        if pd.notna(row["NAME"]):
            definition = row["NAME"]
        elif pd.notna(row["GENE_DESC"]):
            definition = row["GENE_DESC"]
        else:
            definition = None

        term = Term(
            reference=Reference(prefix=PREFIX, identifier=row["GENE_RGD_ID"], name=row["SYMBOL"]),
            definition=definition,
        )
        old_names = row["OLD_NAME"]
        if old_names and pd.notna(old_names):
            for old_name in old_names.split(";"):
                term.append_synonym(Synonym(name=old_name, type=old_name_type))
        old_symbols = row["OLD_SYMBOL"]
        if old_symbols and pd.notna(old_symbols):
            for old_symbol in old_symbols.split(";"):
                term.append_synonym(Synonym(name=old_symbol, type=old_symbol_type))
        for prefix, key in namespace_to_column:
            xref_ids = str(row[key])
            if xref_ids and pd.notna(xref_ids):
                for xref_id in xref_ids.split(";"):
                    if xref_id == "nan":
                        continue
                    if prefix == "uniprot":
                        term.append_relationship(
                            has_gene_product, Reference.auto(prefix=prefix, identifier=xref_id)
                        )
                    elif prefix == "ensembl":
                        if xref_id.startswith("ENSMUSG") or xref_id.startswith("ENSRNOG"):
                            # second one is reverse strand
                            term.append_xref(Reference(prefix=prefix, identifier=xref_id))
                        elif xref_id.startswith("ENSMUST"):
                            term.append_relationship(
                                transcribes_to, Reference(prefix=prefix, identifier=xref_id)
                            )
                        elif xref_id.startswith("ENSMUSP"):
                            term.append_relationship(
                                has_gene_product, Reference(prefix=prefix, identifier=xref_id)
                            )
                        else:
                            logger.warning("[%s] unhandled xref ensembl:%s", PREFIX, xref_id)
                    else:
                        term.append_xref(Reference(prefix=prefix, identifier=xref_id))

        pubmed_ids = row["CURATED_REF_PUBMED_ID"]
        if pubmed_ids and pd.notna(pubmed_ids):
            for pubmed_id in str(pubmed_ids).split(";"):
                term.append_provenance(Reference(prefix="pubmed", identifier=pubmed_id))

        term.set_species(identifier="10116", name="Rattus norvegicus")
        yield term


if __name__ == "__main__":
    RGDGetter.cli()



# -*- coding: utf-8 -*-

"""Get the CCLE Cells, provided by cBioPortal."""

import tarfile
from pathlib import Path
from typing import Iterable, Optional

import pandas as pd
import pystow

from pyobo import Obo, Reference, Term

__all__ = [
    "get_obo",
    "CCLEGetter",
]

PREFIX = "ccle"
VERSION = "2019"


class CCLEGetter(Obo):
    """An ontology representation of the Cancer Cell Line Encyclopedia's cell lines."""

    ontology = bioregistry_key = PREFIX

    def __post_init__(self):  # noqa: D105
        self.data_version = VERSION

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(*, force: bool = False) -> Obo:
    """Get CCLE Cells as OBO."""
    return CCLEGetter(force=force)


def iter_terms(version: Optional[str] = None, force: bool = False) -> Iterable[Term]:
    """Iterate over CCLE Cells."""
    df = ensure_df(version=version, force=force)
    for identifier, depmap_id, name in df.values:
        if pd.isna(name) or pd.isnull(name):
            name = None
        term = Term.from_triple(PREFIX, identifier, name)
        if pd.notna(depmap_id):
            term.append_xref(Reference(prefix="depmap", identifier=depmap_id))
        yield term


def get_version() -> str:
    """Get the default version of CCLE's cell lines."""
    return "2019"


def get_url(version: Optional[str] = None) -> str:
    """Get the cBioPortal URL for the given version of CCLE's cell lines."""
    if version is None:
        version = get_version()
    return f"https://cbioportal-datahub.s3.amazonaws.com/ccle_broad_{version}.tar.gz"


def get_inner(version: Optional[str] = None) -> str:
    """Get the inner tarfile path."""
    if version is None:
        version = get_version()
    return f"ccle_broad_{version}/data_clinical_sample.txt"


def ensure(version: Optional[str] = None, **kwargs) -> Path:
    """Ensure the given version is downloaded."""
    if version is None:
        version = get_version()
    url = get_url(version=version)
    return pystow.ensure("pyobo", "raw", PREFIX, version, url=url, **kwargs)


def ensure_df(version: Optional[str] = None, force: bool = False) -> pd.DataFrame:
    """Get the CCLE clinical sample dataframe."""
    if version is None:
        version = get_version()
    path = ensure(version=version, force=force)
    inner_path = get_inner(version=version)
    with tarfile.open(path) as tf:
        return pd.read_csv(
            tf.extractfile(inner_path),
            sep="\t",
            skiprows=4,  # includes skipping header
            dtype=str,
            usecols=[
                0,  # Sample Identifier
                5,  # DepMap ID
                6,  # Name
                # There are lots of other wonderful sample metadata in case we want more later
            ],
        )


if __name__ == "__main__":
    CCLEGetter.cli()



# -*- coding: utf-8 -*-

"""High-level API for relations."""

import logging
import os
from functools import lru_cache
from typing import List, Mapping, Optional

import pandas as pd

from .utils import get_version
from ..constants import (
    RELATION_COLUMNS,
    RELATION_ID,
    RELATION_PREFIX,
    SOURCE_ID,
    SOURCE_PREFIX,
    TARGET_ID,
    TARGET_PREFIX,
)
from ..getters import get_ontology
from ..identifier_utils import wrap_norm_prefix
from ..struct import Reference, RelationHint, TypeDef, get_reference_tuple
from ..utils.cache import cached_df
from ..utils.path import prefix_cache_join

# TODO get_relation, get_relations

logger = logging.getLogger(__name__)


@wrap_norm_prefix
def get_relations_df(
    prefix: str,
    *,
    use_tqdm: bool = False,
    force: bool = False,
    wide: bool = False,
    strict: bool = True,
) -> pd.DataFrame:
    """Get all relations from the OBO."""
    version = get_version(prefix)
    path = prefix_cache_join(prefix, name="relations.tsv", version=version)

    @cached_df(path=path, dtype=str, force=force)
    def _df_getter() -> pd.DataFrame:
        if force:
            logger.info("[%s] forcing reload for relations", prefix)
        else:
            logger.info("[%s] no cached relations found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, version=version, strict=strict)
        return ontology.get_relations_df(use_tqdm=use_tqdm)

    rv = _df_getter()

    if wide:
        rv = rv.rename(columns={f"{prefix}_id": SOURCE_ID})
        rv[SOURCE_PREFIX] = prefix
        rv = rv[RELATION_COLUMNS]

    return rv


@wrap_norm_prefix
def get_filtered_relations_df(
    prefix: str,
    relation: RelationHint,
    *,
    use_tqdm: bool = False,
    force: bool = False,
    version: Optional[str] = None,
) -> pd.DataFrame:
    """Get all of the given relation."""
    relation_prefix, relation_identifier = relation = get_reference_tuple(relation)
    if version is None:
        version = get_version(prefix)
    path = prefix_cache_join(
        prefix,
        "relations",
        name=f"{relation_prefix}:{relation_identifier}.tsv",
        version=version,
    )
    all_relations_path = prefix_cache_join(prefix, name="relations.tsv", version=version)

    @cached_df(path=path, dtype=str, force=force)
    def _df_getter() -> pd.DataFrame:
        if os.path.exists(all_relations_path):
            logger.debug("[%] loading all relations from %s", prefix, all_relations_path)
            df = pd.read_csv(all_relations_path, sep="\t", dtype=str)
            idx = (df[RELATION_PREFIX] == relation_prefix) & (
                df[RELATION_ID] == relation_identifier
            )
            columns = [f"{prefix}_id", TARGET_PREFIX, TARGET_ID]
            return df.loc[idx, columns]

        logger.info("[%s] no cached relations found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, version=version)
        return ontology.get_filtered_relations_df(relation, use_tqdm=use_tqdm)

    return _df_getter()


@wrap_norm_prefix
def get_id_multirelations_mapping(
    prefix: str,
    typedef: TypeDef,
    *,
    use_tqdm: bool = False,
    force: bool = False,
) -> Mapping[str, List[Reference]]:
    """Get the OBO file and output a synonym dictionary."""
    version = get_version(prefix)
    ontology = get_ontology(prefix, force=force, version=version)
    return ontology.get_id_multirelations_mapping(typedef=typedef, use_tqdm=use_tqdm)


@lru_cache()
@wrap_norm_prefix
def get_relation_mapping(
    prefix: str,
    relation: RelationHint,
    target_prefix: str,
    *,
    use_tqdm: bool = False,
    force: bool = False,
) -> Mapping[str, str]:
    """Get relations from identifiers in the source prefix to target prefix with the given relation.

    .. warning:: Assumes there's only one version of the property for each term.

     Example usage: get homology between HGNC and MGI:

    >>> import pyobo
    >>> human_mapt_hgnc_id = '6893'
    >>> mouse_mapt_mgi_id = '97180'
    >>> hgnc_mgi_orthology_mapping = pyobo.get_relation_mapping('hgnc', 'ro:HOM0000017', 'mgi')
    >>> assert mouse_mapt_mgi_id == hgnc_mgi_orthology_mapping[human_mapt_hgnc_id]
    """
    version = get_version(prefix)
    ontology = get_ontology(prefix, force=force, version=version)
    return ontology.get_relation_mapping(
        relation=relation, target_prefix=target_prefix, use_tqdm=use_tqdm
    )


@wrap_norm_prefix
def get_relation(
    prefix: str,
    source_identifier: str,
    relation: RelationHint,
    target_prefix: str,
    *,
    use_tqdm: bool = False,
    force: bool = False,
) -> Optional[str]:
    """Get the target identifier corresponding to the given relationship from the source prefix/identifier pair.

    .. warning:: Assumes there's only one version of the property for each term.

     Example usage: get homology between MAPT in HGNC and MGI:

    >>> import pyobo
    >>> human_mapt_hgnc_id = '6893'
    >>> mouse_mapt_mgi_id = '97180'
    >>> assert mouse_mapt_mgi_id == pyobo.get_relation('hgnc', human_mapt_hgnc_id, 'ro:HOM0000017', 'mgi')
    """
    relation_mapping = get_relation_mapping(
        prefix=prefix,
        relation=relation,
        target_prefix=target_prefix,
        use_tqdm=use_tqdm,
        force=force,
    )
    return relation_mapping.get(source_identifier)



# -*- coding: utf-8 -*-

"""Converter for SGD."""

from typing import Iterable
from urllib.parse import unquote_plus

from ..struct import Obo, Reference, Synonym, SynonymTypeDef, Term, from_species
from ..utils.path import ensure_tar_df

__all__ = [
    "SGDGetter",
]

HEADER = ["chromosome", "database", "feature", "start", "end", "a", "b", "c", "data"]
PREFIX = "sgd"

URL = (
    "https://downloads.yeastgenome.org/sequence/"
    "S288C_reference/genome_releases/S288C_reference_genome_R64-2-1_20150113.tgz"
)
INNER_PATH = "S288C_reference_genome_R64-2-1_20150113/saccharomyces_cerevisiae_R64-2-1_20150113.gff"

alias_type = SynonymTypeDef.from_text("alias")


class SGDGetter(Obo):
    """An ontology representation of SGD's yeast gene nomenclature."""

    bioversions_key = ontology = PREFIX
    typedefs = [from_species]
    synonym_typedefs = [alias_type]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms for SGD."""
        yield from get_terms(self, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get SGD as OBO."""
    return SGDGetter(force=force)


def get_terms(ontology: Obo, force: bool = False) -> Iterable[Term]:
    """Get SGD terms."""
    df = ensure_tar_df(
        prefix=PREFIX,
        url=URL,
        inner_path=INNER_PATH,
        sep="\t",
        skiprows=18,
        header=None,
        names=HEADER,
        force=force,
        dtype=str,
        version=ontology._version_or_raise,
    )
    df = df[df["feature"] == "gene"]
    for data in df["data"]:
        d = dict(entry.split("=") for entry in data.split(";"))

        identifier = d["dbxref"][len("SGD:") :]
        name = d["Name"]
        definition = unquote_plus(d["Note"])

        synonyms = []

        aliases = d.get("Alias")
        if aliases:
            for alias in aliases.split(","):
                synonyms.append(Synonym(name=unquote_plus(alias), type=alias_type))

        term = Term(
            reference=Reference(prefix=PREFIX, identifier=identifier, name=name),
            definition=definition,
            synonyms=synonyms,
        )
        term.set_species(identifier="4932", name="Saccharomyces cerevisiae")
        yield term


if __name__ == "__main__":
    SGDGetter.cli()



# -*- coding: utf-8 -*-

"""High-level API for properties."""

import logging
import os
from typing import List, Mapping, Optional

import pandas as pd

from .utils import get_version
from ..getters import get_ontology
from ..identifier_utils import wrap_norm_prefix
from ..utils.cache import cached_df, cached_mapping, cached_multidict
from ..utils.io import multidict
from ..utils.path import prefix_cache_join

__all__ = [
    "get_properties_df",
    "get_filtered_properties_df",
    "get_filtered_properties_mapping",
    "get_filtered_properties_multimapping",
    "get_property",
    "get_properties",
]

logger = logging.getLogger(__name__)


@wrap_norm_prefix
def get_properties_df(prefix: str, *, force: bool = False) -> pd.DataFrame:
    """Extract properties.

    :param prefix: the resource to load
    :param force: should the resource be re-downloaded, re-parsed, and re-cached?
    :returns: A dataframe with the properties
    """
    version = get_version(prefix)
    path = prefix_cache_join(prefix, name="properties.tsv", version=version)

    @cached_df(path=path, dtype=str, force=force)
    def _df_getter() -> pd.DataFrame:
        if force:
            logger.info("[%s] forcing reload for properties", prefix)
        else:
            logger.info("[%s] no cached properties found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, version=version)
        df = ontology.get_properties_df()
        df.dropna(inplace=True)
        return df

    return _df_getter()


@wrap_norm_prefix
def get_filtered_properties_mapping(
    prefix: str,
    prop: str,
    *,
    use_tqdm: bool = False,
    force: bool = False,
) -> Mapping[str, str]:
    """Extract a single property for each term as a dictionary.

    :param prefix: the resource to load
    :param prop: the property to extract
    :param use_tqdm: should a progress bar be shown?
    :param force: should the resource be re-downloaded, re-parsed, and re-cached?
    :returns: A mapping from identifier to property value
    """
    version = get_version(prefix)
    path = prefix_cache_join(prefix, "properties", name=f"{prop}.tsv", version=version)
    all_properties_path = prefix_cache_join(prefix, name="properties.tsv", version=version)

    @cached_mapping(path=path, header=[f"{prefix}_id", prop], force=force)
    def _mapping_getter() -> Mapping[str, str]:
        if os.path.exists(all_properties_path):
            logger.info("[%s] loading pre-cached properties", prefix)
            df = pd.read_csv(all_properties_path, sep="\t")
            logger.info("[%s] filtering pre-cached properties", prefix)
            df = df.loc[df["property"] == prop, [f"{prefix}_id", "value"]]
            return dict(df.values)

        logger.info("[%s] no cached properties found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, version=version)
        return ontology.get_filtered_properties_mapping(prop, use_tqdm=use_tqdm)

    return _mapping_getter()


@wrap_norm_prefix
def get_filtered_properties_multimapping(
    prefix: str,
    prop: str,
    *,
    use_tqdm: bool = False,
    force: bool = False,
) -> Mapping[str, List[str]]:
    """Extract multiple properties for each term as a dictionary.

    :param prefix: the resource to load
    :param prop: the property to extract
    :param use_tqdm: should a progress bar be shown?
    :param force: should the resource be re-downloaded, re-parsed, and re-cached?
    :returns: A mapping from identifier to property values
    """
    version = get_version(prefix)
    path = prefix_cache_join(prefix, "properties", name=f"{prop}.tsv", version=version)
    all_properties_path = prefix_cache_join(prefix, name="properties.tsv", version=version)

    @cached_multidict(path=path, header=[f"{prefix}_id", prop], force=force)
    def _mapping_getter() -> Mapping[str, List[str]]:
        if os.path.exists(all_properties_path):
            logger.info("[%s] loading pre-cached properties", prefix)
            df = pd.read_csv(all_properties_path, sep="\t")
            logger.info("[%s] filtering pre-cached properties", prefix)
            df = df.loc[df["property"] == prop, [f"{prefix}_id", "value"]]
            return multidict(df.values)

        logger.info("[%s] no cached properties found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, version=version)
        return ontology.get_filtered_properties_multimapping(prop, use_tqdm=use_tqdm)

    return _mapping_getter()


def get_property(prefix: str, identifier: str, prop: str) -> Optional[str]:
    """Extract a single property for the given entity.

    :param prefix: the resource to load
    :param identifier: the identifier withing the resource
    :param prop: the property to extract
    :returns: The single value for the property. If multiple are expected, use :func:`get_properties`

    >>> import pyobo
    >>> pyobo.get_property('chebi', '132964', 'http://purl.obolibrary.org/obo/chebi/smiles')
    "C1(=CC=C(N=C1)OC2=CC=C(C=C2)O[C@@H](C(OCCCC)=O)C)C(F)(F)F"
    """
    filtered_properties_mapping = get_filtered_properties_mapping(prefix=prefix, prop=prop)
    return filtered_properties_mapping.get(identifier)


def get_properties(prefix: str, identifier: str, prop: str) -> Optional[List[str]]:
    """Extract a set of properties for the given entity.

    :param prefix: the resource to load
    :param identifier: the identifier withing the resource
    :param prop: the property to extract
    :returns: Multiple values for the property. If only one is expected, use :func:`get_property`
    """
    filtered_properties_multimapping = get_filtered_properties_multimapping(
        prefix=prefix, prop=prop
    )
    return filtered_properties_multimapping.get(identifier)


@wrap_norm_prefix
def get_filtered_properties_df(
    prefix: str,
    prop: str,
    *,
    use_tqdm: bool = False,
    force: bool = False,
) -> pd.DataFrame:
    """Extract a single property for each term.

    :param prefix: the resource to load
    :param prop: the property to extract
    :param use_tqdm: should a progress bar be shown?
    :param force: should the resource be re-downloaded, re-parsed, and re-cached?
    :returns: A dataframe from identifier to property value. Columns are [<prefix>_id, value].
    """
    version = get_version(prefix)
    path = prefix_cache_join(prefix, "properties", name=f"{prop}.tsv", version=version)
    all_properties_path = prefix_cache_join(prefix, name="properties.tsv", version=version)

    @cached_df(path=path, dtype=str, force=force)
    def _df_getter() -> pd.DataFrame:
        if os.path.exists(all_properties_path):
            logger.info("[%s] loading pre-cached properties", prefix)
            df = pd.read_csv(all_properties_path, sep="\t")
            logger.info("[%s] filtering pre-cached properties", prefix)
            return df.loc[df["property"] == prop, [f"{prefix}_id", "value"]]

        if force:
            logger.info("[%s] forcing reload for properties", prefix)
        else:
            logger.info("[%s] no cached properties found. getting from OBO loader", prefix)
        ontology = get_ontology(prefix, force=force, version=version)
        return ontology.get_filtered_properties_df(prop, use_tqdm=use_tqdm)

    return _df_getter()



# -*- coding: utf-8 -*-

"""Converter for CPT."""

from typing import Iterable

import pandas as pd

from pyobo import Obo, Reference, Term

__all__ = [
    "CPTGetter",
]

cpt_url = "https://www2a.cdc.gov/vaccines/iis/iisstandards/downloads/cpt.txt"
PREFIX = "cpt"


class CPTGetter(Obo):
    """An ontology representation of CPT."""

    ontology = PREFIX
    dynamic_version = True

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms()


def iter_terms() -> Iterable[Term]:
    """Iterate over CPT terms."""
    df = pd.read_csv(
        cpt_url,
        sep="|",
        names=[
            "cpt",
            "description",
            "...",
            "vaccine_name",
            "cvx",
            "comments",
            "updated",
            "internal_id",
        ],
        dtype=str,
    )
    del df["..."]
    for col in df.columns:
        df[col] = df[col].map(lambda s: s.strip() if pd.notna(s) else s)
    for cpt_id, description, name, cvx, comments, _updated, _internal_id in df.values:
        term = Term(
            reference=Reference(prefix=PREFIX, identifier=cpt_id, name=name),
            definition=description,
        )
        if pd.notna(cvx):
            term.append_xref(Reference(prefix="cvx", identifier=cvx))
        if pd.notna(comments):
            term.append_comment(comments)
        yield term


if __name__ == "__main__":
    CPTGetter.cli()



# -*- coding: utf-8 -*-

"""Gene Ontology."""

from pyobo import get_descendants

__all__ = [
    "is_biological_process",
    "is_molecular_function",
    "is_cellular_component",
]


def is_biological_process(identifier: str) -> bool:
    """Return if the given GO identifier is a biological process.

    >>> is_biological_process('0006915')
    True
    >>> is_biological_process('GO:0006915')
    True
    >>> is_molecular_function('0006915')
    False
    >>> is_cellular_component('0006915')
    False
    """
    return _is_descendant(identifier, "0008150")


def is_molecular_function(identifier: str) -> bool:
    """Return if the given GO identifier is a molecular function."""
    return _is_descendant(identifier, "0003674")


def is_cellular_component(identifier: str) -> bool:
    """Return if the given GO identifier is a cellular component."""
    return _is_descendant(identifier, "0005575")


def _is_descendant(identifier: str, ancestor: str) -> bool:
    identifier = identifier.lower()
    if not identifier.startswith("go:"):
        identifier = f"go:{identifier}"
    descendants = get_descendants("go", ancestor)
    return descendants is not None and identifier in descendants


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-

"""Converter for UniProt."""

from pathlib import Path
from typing import Iterable, Optional

import bioversions
from tqdm.auto import tqdm

from pyobo import Obo, Reference
from pyobo.constants import RAW_MODULE
from pyobo.identifier_utils import standardize_ec
from pyobo.struct import Term, enables, from_species
from pyobo.utils.io import open_reader

PREFIX = "uniprot"
REVIEWED_URL = (
    "https://rest.uniprot.org/uniprotkb/stream?compressed=true"
    "&fields=accession%2Cid%2Corganism_id%2Cprotein_name%2Cec%2Clit_pubmed_id%2Cxref_pdb"
    "&format=tsv&query=%28%2A%29%20AND%20%28reviewed%3Atrue%29"
)


class UniProtGetter(Obo):
    """An ontology representation of the UniProt database."""

    bioversions_key = ontology = PREFIX
    typedefs = [from_species, enables]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        yield from iter_terms(force=force, version=self._version_or_raise)


def get_obo(force: bool = False) -> Obo:
    """Get UniProt as OBO."""
    return UniProtGetter(force=force)


def iter_terms(version: Optional[str] = None, force: bool = False) -> Iterable[Term]:
    """Iterate over UniProt Terms."""
    with open_reader(ensure(version=version, force=force)) as reader:
        _ = next(reader)  # header
        for uniprot_id, name, taxonomy_id, _synonyms, ecs, pubmeds, pdbs in tqdm(
            reader, desc="Mapping UniProt", unit_scale=True
        ):
            term = Term.from_triple(prefix=PREFIX, identifier=uniprot_id, name=name)
            # TODO add gene encodes from relationship
            # TODO add description
            term.set_species(taxonomy_id)
            if ecs:
                for ec in ecs.split(";"):
                    term.append_relationship(
                        enables, Reference(prefix="eccode", identifier=standardize_ec(ec))
                    )
            for pubmed in pubmeds.split(";"):
                if pubmed:
                    term.append_provenance(Reference(prefix="pubmed", identifier=pubmed.strip()))
            for pdb in pdbs.split(";"):
                if pdb:
                    term.append_xref(Reference(prefix="pdb", identifier=pdb.strip()))
            yield term


def ensure(version: Optional[str] = None, force: bool = False) -> Path:
    """Ensure the reviewed uniprot names are available."""
    if version is None:
        version = bioversions.get_version("uniprot")
    return RAW_MODULE.ensure(PREFIX, version, name="reviewed.tsv.gz", url=REVIEWED_URL, force=force)


if __name__ == "__main__":
    UniProtGetter.cli()



# -*- coding: utf-8 -*-

"""Get DrugCentral as OBO."""

import logging
from collections import defaultdict
from contextlib import closing
from typing import DefaultDict, Iterable, List

import bioregistry
import psycopg2
from tqdm.auto import tqdm

from pyobo.struct import Obo, Reference, Synonym, Term

__all__ = [
    "DrugCentralGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "drugcentral"

HOST = "unmtid-dbs.net"
PORT = 5433
USER = "drugman"
PASSWORD = "dosage"
DBNAME = "drugcentral"
PARAMS = dict(dbname=DBNAME, user=USER, password=PASSWORD, host=HOST, port=PORT)


class DrugCentralGetter(Obo):
    """An ontology representation of the DrugCentral database."""

    ontology = bioversions_key = PREFIX

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms()


def get_obo(force: bool = False) -> Obo:
    """Get DrugCentral OBO."""
    return DrugCentralGetter(force=force)


def iter_terms() -> Iterable[Term]:
    """Iterate over DrugCentral terms."""
    with closing(psycopg2.connect(**PARAMS)) as conn:
        with closing(conn.cursor()) as cur:
            cur.execute(
                "SELECT cd_id, name, cas_reg_no, mrdef, inchi, smiles, inchikey FROM public.structures"
            )
            structures = cur.fetchall()

        with closing(conn.cursor()) as cur:
            cur.execute("SELECT struct_id, id_type, identifier FROM public.identifier")
            rows = cur.fetchall()
            xrefs: DefaultDict[str, List[Reference]] = defaultdict(list)
            for drugcentral_id, prefix, identifier in tqdm(
                rows, unit_scale=True, desc="loading xrefs"
            ):
                if not identifier or not prefix:
                    continue
                if prefix == "ChEMBL_ID":
                    prefix = "chembl.compound"
                xref_prefix_norm = bioregistry.normalize_prefix(prefix)
                if xref_prefix_norm is None:
                    tqdm.write(f"did not normalize {prefix}:{identifier}")
                    continue
                identifier = bioregistry.standardize_identifier(xref_prefix_norm, identifier)
                xrefs[str(drugcentral_id)].append(
                    Reference(prefix=xref_prefix_norm, identifier=identifier)
                )
        with closing(conn.cursor()) as cur:
            cur.execute("SELECT id, name FROM public.synonyms")
            synonyms: DefaultDict[str, List[Synonym]] = defaultdict(list)
            for drugcentral_id, synonym in cur.fetchall():
                synonyms[str(drugcentral_id)].append(Synonym(name=synonym))

    for drugcentral_id, name, cas, definition, inchi, smiles, inchi_key in structures:
        drugcentral_id = str(drugcentral_id)
        term = Term(
            reference=Reference(prefix=PREFIX, identifier=drugcentral_id, name=name),
            definition=definition,
            synonyms=synonyms.get(drugcentral_id, []),
            xrefs=xrefs.get(drugcentral_id, []),
        )
        if inchi_key:
            term.append_xref(Reference(prefix="inchikey", identifier=inchi_key))
        term.append_property("smiles", smiles)
        term.append_property("inchi", inchi)
        if cas:
            term.append_xref(Reference(prefix="cas", identifier=cas))
        yield term


if __name__ == "__main__":
    DrugCentralGetter.cli()



# -*- coding: utf-8 -*-

"""Selventa complexes."""

from typing import Iterable

import pandas as pd

from pyobo import Obo, Term
from pyobo.utils.path import ensure_df

__all__ = [
    "SCOMPGetter",
]

PREFIX = "scomp"
URL = "https://raw.githubusercontent.com/OpenBEL/resource-generator/master/datasets/selventa-named-complexes.txt"


class SCOMPGetter(Obo):
    """An ontology representation of the Selventa protein complex nomenclature."""

    ontology = PREFIX
    static_version = "1.0.0"

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(force=force)


def get_obo(*, force: bool = False) -> Obo:
    """Get Selventa Complexes as OBO."""
    return SCOMPGetter(force=force)


def iter_terms(force: bool = False) -> Iterable[Term]:
    """Iterate over selventa complex terms."""
    df = ensure_df(PREFIX, url=URL, skiprows=9, force=force)

    terms = {}
    for identifier, label, synonyms, xref in df[["ID", "LABEL", "SYNONYMS", "XREF"]].values:
        term = Term.from_triple(PREFIX, identifier, label)
        for synonym in synonyms.split("|") if pd.notna(synonyms) else []:
            term.append_synonym(synonym)
        if pd.notna(xref):
            term.append_xref(xref)
        terms[identifier] = term

    df.PARENTS = df.PARENTS.map(lambda x: x[len("SCOMP:") :], na_action="ignore")
    for child, parent in df.loc[df.PARENTS.notna(), ["ID", "PARENTS"]].values:
        if child == parent:
            continue  # wow...
        terms[child].append_parent(terms[parent])

    yield from terms.values()


if __name__ == "__main__":
    get_obo().write_default(write_obo=True, force=True)



# -*- coding: utf-8 -*-

"""Converter for miRBase Families."""

from typing import Iterable

import pandas as pd
from tqdm.auto import tqdm

from .mirbase_constants import (
    get_premature_df,
    get_premature_family_df,
    get_premature_to_prefamily_df,
)
from ..struct import Obo, Reference, Term, has_member

__all__ = [
    "MiRBaseFamilyGetter",
]

PREFIX = "mirbase.family"


class MiRBaseFamilyGetter(Obo):
    """An ontology representation of miRBase's miRNA family nomenclature."""

    ontology = PREFIX
    bioversions_key = "mirbase"

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get miRBase family as OBO."""
    return MiRBaseFamilyGetter(force=force)


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Get miRBase family terms."""
    df = get_df(version, force=force)
    for family_id, name, mirna_id, mirna_name in tqdm(df.values, total=len(df.index)):
        term = Term(
            reference=Reference(prefix=PREFIX, identifier=family_id, name=name),
        )
        term.append_relationship(
            has_member, Reference(prefix="mirna", identifier=mirna_id, name=mirna_name)
        )
        yield term


def get_df(version: str, force: bool = False) -> pd.DataFrame:
    """Get the miRBase family dataframe."""
    mirna_prefamily_df = get_premature_to_prefamily_df(version, force=force)
    prefamily_df = get_premature_family_df(version, force=force)
    premature_df = get_premature_df(version, force=force)
    intermediate_df = pd.merge(
        mirna_prefamily_df, prefamily_df, left_on="prefamily_key", right_on="prefamily_key"
    )
    rv = pd.merge(intermediate_df, premature_df, left_on="premature_key", right_on="premature_key")
    del rv["premature_key"]
    del rv["prefamily_key"]
    return rv


if __name__ == "__main__":
    get_obo().write_default(use_tqdm=True)



# -*- coding: utf-8 -*-

"""Default typedefs, references, and other structures."""

from dataclasses import dataclass, field
from typing import Dict, Iterable, List, Optional, Tuple, Union

from .reference import Reference, Referenced
from ..identifier_utils import normalize_curie
from ..resources.ro import load_ro

__all__ = [
    "TypeDef",
    "RelationHint",
    "get_reference_tuple",
    "default_typedefs",
    "from_species",
    "species_specific",
    "has_part",
    "part_of",
    "is_a",
    "see_also",
    "has_member",
    "member_of",
    "superclass_of",
    "orthologous",
    "has_role",
    "role_of",
    "has_mature",
    "has_gene_product",
    "gene_product_member_of",
    "has_gene_product",
    "transcribes_to",
    "translates_to",
    "gene_product_member_of",
    "example_of_usage",
    "alternative_term",
    "editor_note",
    "has_salt",
    "enables",
    "participates_in",
    "has_participant",
]


@dataclass
class TypeDef(Referenced):
    """A type definition in OBO."""

    reference: Reference
    comment: Optional[str] = None
    namespace: Optional[str] = None
    definition: Optional[str] = None
    is_transitive: Optional[bool] = None
    is_symmetric: Optional[bool] = None
    domain: Optional[Reference] = None
    range: Optional[Reference] = None
    parents: List[Reference] = field(default_factory=list)
    xrefs: List[Reference] = field(default_factory=list)
    inverse: Optional[Reference] = None
    created_by: Optional[str] = None
    holds_over_chain: Optional[List[Reference]] = None
    #: Whether this relationship is a metadata tag. Properties that are marked as metadata tags are
    #: used to record object metadata. Object metadata is additional information about an object
    #: that is useful to track, but does not impact the definition of the object or how it should
    #: be treated by a reasoner. Metadata tags might be used to record special term synonyms or
    #: structured notes about a term, for example.
    is_metadata_tag: Optional[bool] = None

    def __hash__(self) -> int:  # noqa: D105
        return hash((self.__class__, self.prefix, self.identifier))

    def iterate_obo_lines(self) -> Iterable[str]:
        """Iterate over the lines to write in an OBO file."""
        yield "\n[Typedef]"
        yield f"id: {self.reference.curie}"
        if self.name:
            yield f"name: {self.reference.name}"
        if self.definition:
            yield f'def: "{self.definition}"'

        if self.is_metadata_tag is not None:
            yield f'is_metadata_tag: {"true" if self.is_metadata_tag else "false"}'

        if self.namespace:
            yield f"namespace: {self.namespace}"

        if self.created_by:
            yield f"created_by: {self.created_by}"

        if self.comment:
            yield f"comment: {self.comment}"

        for xref in self.xrefs:
            yield f"xref: {xref}"

        if self.is_transitive is not None:
            yield f'is_transitive: {"true" if self.is_transitive else "false"}'

        if self.is_symmetric is not None:
            yield f'is_symmetric: {"true" if self.is_symmetric else "false"}'
        if self.holds_over_chain:
            _chain = " ".join(link.curie for link in self.holds_over_chain)
            _names = " / ".join(link.name or "_" for link in self.holds_over_chain)
            yield f"holds_over_chain: {_chain} ! {_names}"

    @classmethod
    def from_triple(cls, prefix: str, identifier: str, name: Optional[str] = None) -> "TypeDef":
        """Create a typedef from a reference."""
        return cls(reference=Reference(prefix=prefix, identifier=identifier, name=name))

    @classmethod
    def from_curie(cls, curie: str, name: Optional[str] = None) -> "TypeDef":
        """Create a TypeDef directly from a CURIE and optional name."""
        prefix, identifier = normalize_curie(curie)
        if prefix is None or identifier is None:
            raise ValueError
        return cls.from_triple(prefix=prefix, identifier=identifier, name=name)


RelationHint = Union[Reference, TypeDef, Tuple[str, str], str]


def get_reference_tuple(relation: RelationHint) -> Tuple[str, str]:
    """Get tuple for typedef/reference."""
    if isinstance(relation, (Reference, TypeDef)):
        return relation.prefix, relation.identifier
    elif isinstance(relation, tuple):
        return relation
    elif isinstance(relation, str):
        prefix, identifier = normalize_curie(relation)
        if prefix is None or identifier is None:
            raise ValueError(f"string given is not valid curie: {relation}")
        return prefix, identifier
    else:
        raise TypeError(f"Relation is invalid type: {relation}")


RO_PREFIX = "RO"
BFO_PREFIX = "BFO"
IAO_PREFIX = "IAO"
SIO_PREFIX = "SIO"

from_species = TypeDef(
    reference=Reference(prefix=RO_PREFIX, identifier="0002162", name="in taxon"),
)
species_specific = TypeDef(
    reference=Reference(prefix="debio", identifier="0000007", name="species specific"),
    definition="X speciesSpecific Y means that Y is a general phenomena, "
    "like a pathway, and X is the version that appears in a species. X should state which"
    "species with RO:0002162 (in taxon)",
)

part_of = TypeDef(
    reference=Reference(prefix=BFO_PREFIX, identifier="0000050", name="part of"),
    comment="Inverse of has_part",
    inverse=Reference(prefix=BFO_PREFIX, identifier="0000051", name="has part"),
)
has_part = TypeDef(
    reference=Reference(prefix=BFO_PREFIX, identifier="0000051", name="has part"),
    comment="Inverse of part_of",
    inverse=Reference(prefix=BFO_PREFIX, identifier="0000050", name="part of"),
)
participates_in = TypeDef(
    reference=Reference(prefix=RO_PREFIX, identifier="0000056", name="participates in"),
    comment="Inverse of has participant",
    inverse=Reference(prefix=RO_PREFIX, identifier="0000057", name="has participant"),
)
has_participant = TypeDef(
    reference=Reference(prefix=RO_PREFIX, identifier="0000057", name="has participant"),
    comment="Inverse of has participant",
    inverse=Reference(prefix=RO_PREFIX, identifier="0000056", name="participates in"),
)
exact_match = TypeDef(
    reference=Reference(prefix="skos", identifier="exactMatch", name="exact match"),
)
is_a = TypeDef(
    reference=Reference(prefix="rdfs", identifier="subClassOf", name="subclass of"),
)
see_also = TypeDef(
    reference=Reference(prefix="rdfs", identifier="seeAlso", name="see also"),
)
comment = TypeDef(reference=Reference(prefix="rdfs", identifier="comment", name="comment"))
has_member = TypeDef(
    reference=Reference(prefix=RO_PREFIX, identifier="0002351", name="has member"),
)
member_of = TypeDef(
    reference=Reference(prefix=RO_PREFIX, identifier="0002350", name="member of"),
    inverse=has_member.reference,
)
superclass_of = TypeDef(
    reference=Reference(prefix="sssom", identifier="superClassOf", name="super class of"),
    comment="Inverse of subClassOf",
    inverse=is_a.reference,
)

develops_from = TypeDef.from_triple(prefix=RO_PREFIX, identifier="0002202", name="develops from")
orthologous = TypeDef(
    reference=Reference(
        prefix=RO_PREFIX, identifier="HOM0000017", name="in orthology relationship with"
    ),
    is_symmetric=True,
)

has_role = TypeDef(
    reference=Reference(prefix=RO_PREFIX, identifier="0000087", name="has role"),
    definition="a relation between an independent continuant (the bearer) and a role,"
    " in which the role specifically depends on the bearer for its existence",
    domain=Reference(prefix=BFO_PREFIX, identifier="0000004", name="independent continuant"),
    range=Reference(prefix=BFO_PREFIX, identifier="0000023", name="role"),
    parents=[Reference(prefix=RO_PREFIX, identifier="0000053", name="bearer of")],
    inverse=Reference(prefix=RO_PREFIX, identifier="0000081", name="role of"),
)

role_of = TypeDef(
    reference=Reference(prefix=RO_PREFIX, identifier="0000081", name="role of"),
    definition="a relation between a role and an independent continuant (the bearer),"
    " in which the role specifically depends on the bearer for its existence",
    parents=[Reference(prefix=RO_PREFIX, identifier="0000052", name="inheres in")],
    inverse=has_role.reference,
)

has_mature = TypeDef(
    reference=Reference(prefix="debio", identifier="0000002", name="has mature miRNA"),
)

transcribes_to = TypeDef(
    reference=Reference(prefix=RO_PREFIX, identifier="0002511", name="transcribed to"),
)
translates_to = TypeDef(
    reference=Reference(prefix=RO_PREFIX, identifier="0002513", name="ribosomally translates to"),
)
gene_product_of = TypeDef.from_triple(
    prefix=RO_PREFIX, identifier="0002204", name="gene product of"
)
has_gene_product = TypeDef(
    reference=Reference(prefix=RO_PREFIX, identifier="0002205", name="has gene product"),
    inverse=gene_product_of.reference,
)  # holds over chain (transcribes_to, translates_to)
gene_product_member_of = TypeDef(
    reference=Reference(prefix="debio", identifier="0000001", name="gene product is a member of"),
    holds_over_chain=[
        has_gene_product.reference,
        member_of.reference,
    ],
)

has_salt = TypeDef(
    reference=Reference(prefix="debio", identifier="0000006", name="has salt"),
)

example_of_usage = Reference(prefix=IAO_PREFIX, identifier="0000112", name="example of usage")
alternative_term = Reference(prefix=IAO_PREFIX, identifier="0000118", name="alternative term")
editor_note = Reference(prefix=IAO_PREFIX, identifier="0000116", name="editor note")

is_immediately_transformed_from = TypeDef.from_triple(
    prefix=SIO_PREFIX, identifier="000658", name="is immediately transformed from"
)
enables = TypeDef.from_triple(prefix="RO", identifier="0002327", name="enables")

"""ChEBI"""

is_conjugate_base_of = TypeDef(
    reference=Reference(
        prefix="chebi", identifier="is_conjugate_base_of", name="is conjugate base of"
    ),
)
is_conjugate_acid_of = TypeDef(
    reference=Reference(
        prefix="chebi", identifier="is_conjugate_acid_of", name="is conjugate acid of"
    ),
)
is_enantiomer_of = TypeDef(
    reference=Reference(prefix="chebi", identifier="is_enantiomer_of", name="is enantiomer of"),
)
is_tautomer_of = TypeDef(
    reference=Reference(prefix="chebi", identifier="is_tautomer_of", name="is tautomer of"),
)
has_parent_hydride = TypeDef(
    reference=Reference(prefix="chebi", identifier="has_parent_hydride", name="has parent hydride"),
)
is_substituent_group_from = TypeDef(
    reference=Reference(
        prefix="chebi", identifier="is_substituent_group_from", name="is substituent group from"
    ),
)
has_functional_parent = TypeDef(
    reference=Reference(
        prefix="chebi", identifier="has_functional_parent", name="has functional parent"
    ),
)

default_typedefs: Dict[Tuple[str, str], TypeDef] = {
    v.pair: v for k, v in locals().items() if isinstance(v, TypeDef)
}

for pair, name in load_ro().items():
    if pair not in default_typedefs:
        default_typedefs[pair] = TypeDef.from_triple(pair[0], pair[1], name)



# -*- coding: utf-8 -*-

"""Get the Biomappings manually curated equivalences."""

import pandas as pd
from pystow.utils import get_commit

from pyobo.constants import (
    PROVENANCE,
    SOURCE_ID,
    SOURCE_PREFIX,
    TARGET_ID,
    TARGET_PREFIX,
    XREF_COLUMNS,
)

__all__ = [
    "get_biomappings_df",
]


def get_biomappings_df() -> pd.DataFrame:
    """Get biomappings equivalences."""
    sha = get_commit("biopragmatics", "biomappings")
    url = f"https://raw.githubusercontent.com/biopragmatics/biomappings/{sha}/src/biomappings/resources/mappings.tsv"
    df = pd.read_csv(url, sep="\t")
    df[PROVENANCE] = url
    df.rename(
        columns={
            "source prefix": SOURCE_PREFIX,
            "source identifier": SOURCE_ID,
            "target prefix": TARGET_PREFIX,
            "target identifier": TARGET_ID,
        },
        inplace=True,
    )
    df = df[XREF_COLUMNS]
    return df



# -*- coding: utf-8 -*-

"""Converter for UMLS.

Run with ``python -m pyobo.sources.umls``
"""

import itertools as itt
import operator
import os
from typing import Iterable

import bioregistry
import pandas as pd
from tqdm.auto import tqdm
from umls_downloader import open_umls

from pyobo import Obo, Reference, Synonym, SynonymTypeDef, Term
from pyobo.utils.io import open_map_tsv

__all__ = [
    "UMLSGetter",
]

HERE = os.path.abspath(os.path.dirname(__file__))
SYNONYM_TYPE_PATH = os.path.join(HERE, "synonym_types.tsv")

RRF_COLUMNS = [
    "CUI",
    "LAT - Language",
    "TS - Term Status",
    "LUI - Local Unique Identifier",
    "STT - String Type",
    "SUI - Unique Identifier for String",
    "ISPREF - is preferred",
    "AUI - Unique atom identifier",
    "SAUI - Source atom identifier",
    "SCUI - Source concept identifier",
    "SDUI - Source descriptor identifier",
    "SAB - source name",
    "TTY - Term Type in Source",
    "CODE",
    "STR",
    "SRL",
    "SUPPRESS",
    "CVF",
    "?",
]

PREFIX = "umls"

SOURCE_VOCAB_URL = "https://www.nlm.nih.gov/research/umls/sourcereleasedocs/index.html"

SYNONYM_ABB = open_map_tsv(SYNONYM_TYPE_PATH)


class UMLSGetter(Obo):
    """An ontology representation of UMLS."""

    ontology = bioversions_key = PREFIX
    synonym_typedefs = [SynonymTypeDef.from_text(v) for v in SYNONYM_ABB.values()]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise)


def get_obo() -> Obo:
    """Get UMLS as OBO."""
    return UMLSGetter()


def iter_terms(version: str) -> Iterable[Term]:
    """Iterate over UMLS terms."""
    with open_umls(version=version) as file:
        it = tqdm(file, unit_scale=True, desc="[umls] parsing")
        lines = (line.decode("utf-8").strip().split("|") for line in it)
        for cui, cui_lines in itt.groupby(lines, key=operator.itemgetter(0)):
            df = pd.DataFrame(list(cui_lines), columns=RRF_COLUMNS)
            df = df[df["LAT - Language"] == "ENG"]
            idx = (
                (df["ISPREF - is preferred"] == "Y")
                & (df["TS - Term Status"] == "P")
                & (df["STT - String Type"] == "PF"),
            )
            pref_rows_df = df.loc[idx]
            if len(pref_rows_df.index) != 1:
                it.write(f"no preferred term for umls:{cui}. got {len(pref_rows_df.index)}")
                continue

            df["TTY - Term Type in Source"] = df["TTY - Term Type in Source"].map(
                SYNONYM_ABB.__getitem__
            )

            _r = pref_rows_df.iloc[0]
            sdf = df[["SAB - source name", "CODE", "TTY - Term Type in Source", "STR"]]

            synonyms = []
            xrefs = []
            for source, identifier, synonym_type, synonym in sdf.values:
                norm_source = bioregistry.normalize_prefix(source)
                if norm_source is None or not identifier:
                    provenance = []
                else:
                    ref = Reference(prefix=norm_source, identifier=identifier)
                    provenance = [ref]
                    xrefs.append(ref)
                synonyms.append(
                    Synonym(
                        name=synonym,
                        provenance=provenance,
                        type=SynonymTypeDef.from_text(synonym_type),
                    )
                )

            xrefs = sorted(
                set(xrefs), key=lambda reference: (reference.prefix, reference.identifier)
            )

            term = Term(
                reference=Reference(prefix=PREFIX, identifier=cui, name=_r["STR"]),
                synonyms=synonyms,
                xrefs=xrefs,
            )
            yield term


if __name__ == "__main__":
    UMLSGetter.cli()



# -*- coding: utf-8 -*-

"""Get FamPlex xrefs."""

import logging
from functools import lru_cache
from typing import Mapping, Tuple

import bioregistry
import pandas as pd

from ...constants import (
    PROVENANCE,
    SOURCE_ID,
    SOURCE_PREFIX,
    TARGET_ID,
    TARGET_PREFIX,
    XREF_COLUMNS,
)
from ...utils.path import ensure_df

__all__ = [
    "get_famplex_xrefs_df",
]

logger = logging.getLogger(__name__)

URL = "https://github.com/sorgerlab/famplex/raw/master/equivalences.csv"


def _get_famplex_df(force: bool = False) -> pd.DataFrame:
    return ensure_df(
        prefix="fplx",
        url=URL,
        force=force,
        header=None,
        names=[TARGET_PREFIX, TARGET_ID, SOURCE_ID],
        sep=",",
    )


def get_famplex_xrefs_df(force: bool = False) -> pd.DataFrame:
    """Get xrefs from FamPlex."""
    df = _get_famplex_df(force=force)
    df[TARGET_PREFIX] = df[TARGET_PREFIX].map(bioregistry.normalize_prefix)
    df = df[df[TARGET_PREFIX].notna()]
    df[SOURCE_PREFIX] = "fplx"
    df[PROVENANCE] = "https://github.com/sorgerlab/famplex/raw/master/equivalences.csv"
    df = df[XREF_COLUMNS]
    return df


@lru_cache()
def get_remapping(force: bool = False) -> Mapping[Tuple[str, str], Tuple[str, str, str]]:
    """Get a mapping from database/identifier pairs to famplex identifiers."""
    df = _get_famplex_df(force=force)
    rv = {}
    for target_ns, target_id, source_id in df.values:
        if target_ns.lower() == "medscan":
            continue  # MEDSCAN is proprietary and Ben said to skip using these identifiers
        remapped_prefix = bioregistry.normalize_prefix(target_ns)
        if remapped_prefix is None:
            logger.warning("could not remap %s", target_ns)
        else:
            rv[remapped_prefix, target_id] = "fplx", source_id, source_id
    return rv



# -*- coding: utf-8 -*-

"""Get the xrefs from IntAct."""

from typing import Mapping

import pandas as pd

from pyobo.api.utils import get_version
from pyobo.constants import PROVENANCE, SOURCE_PREFIX, TARGET_PREFIX, XREF_COLUMNS
from pyobo.utils.cache import cached_mapping
from pyobo.utils.path import prefix_cache_join

__all__ = [
    "COMPLEXPORTAL_MAPPINGS",
    "get_intact_complex_portal_xrefs_df",
    "get_complexportal_mapping",
    "get_intact_reactome_xrefs_df",
    "get_reactome_mapping",
]

COMPLEXPORTAL_MAPPINGS = (
    "ftp://ftp.ebi.ac.uk/pub/databases/intact/current/various/cpx_ebi_ac_translation.txt"
)
REACTOME_MAPPINGS = "ftp://ftp.ebi.ac.uk/pub/databases/intact/current/various/reactome.dat"


def _get_complexportal_df():
    return pd.read_csv(
        COMPLEXPORTAL_MAPPINGS, sep="\t", header=None, names=["source_id", "target_id"]
    )


def get_intact_complex_portal_xrefs_df() -> pd.DataFrame:
    """Get IntAct-Complex Portal xrefs."""
    df = _get_complexportal_df()
    df[SOURCE_PREFIX] = "intact"
    df[TARGET_PREFIX] = "complexportal"
    df[PROVENANCE] = COMPLEXPORTAL_MAPPINGS
    df = df[XREF_COLUMNS]
    return df


def get_complexportal_mapping() -> Mapping[str, str]:
    """Get IntAct to Complex Portal mapping.

    Is basically equivalent to:

    .. code-block:: python

        from pyobo import get_filtered_xrefs
        intact_complexportal_mapping = get_filtered_xrefs('intact', 'complexportal')
    """

    @cached_mapping(
        path=prefix_cache_join(
            "intact", "xrefs", name="complexportal.tsv", version=get_version("intact")
        ),
        header=["intact_id", "complexportal_id"],
    )
    def _cache():
        df = _get_complexportal_df()
        return dict(df.values)

    return _cache()


def _get_reactome_df():
    return pd.read_csv(REACTOME_MAPPINGS, sep="\t", header=None, names=["source_id", "target_id"])


def get_intact_reactome_xrefs_df() -> pd.DataFrame:
    """Get IntAct-Reactome xrefs."""
    df = _get_reactome_df()
    df[SOURCE_PREFIX] = "intact"
    df[TARGET_PREFIX] = "reactome"
    df[PROVENANCE] = REACTOME_MAPPINGS
    df = df[XREF_COLUMNS]
    return df


def get_reactome_mapping() -> Mapping[str, str]:
    """Get IntAct to Reactome mapping.

    Is basically equivalent to:

    .. code-block:: python

        from pyobo import get_filtered_xrefs
        intact_complexportal_mapping = get_filtered_xrefs('intact', 'reactome')
    """

    @cached_mapping(
        path=prefix_cache_join(
            "intact", "xrefs", name="reactome.tsv", version=get_version("intact")
        ),
        header=["intact_id", "reactome_id"],
    )
    def _cache():
        df = _get_complexportal_df()
        return dict(df.values)

    return _cache()


def get_xrefs_df() -> pd.DataFrame:
    """Get IntAct xrefs."""
    return pd.concat(
        [
            get_intact_complex_portal_xrefs_df(),
            get_intact_reactome_xrefs_df(),
        ]
    )



# -*- coding: utf-8 -*-

"""Converter for the Antibody Registry."""

import logging
from typing import Iterable, Mapping, Optional

import bioversions
import pandas as pd
from bioregistry.utils import removeprefix
from tqdm.auto import tqdm

from pyobo import Obo, Term
from pyobo.utils.path import ensure_df

__all__ = [
    "AntibodyRegistryGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "antibodyregistry"
URL = "http://antibodyregistry.org/php/fileHandler.php"
CHUNKSIZE = 20_000


def get_chunks(force: bool = False) -> pd.DataFrame:
    """Get the BioGRID identifiers mapping dataframe."""
    version = bioversions.get_version(PREFIX)
    df = ensure_df(
        PREFIX,
        url=URL,
        name="results.csv",
        force=force,
        version=version,
        sep=",",
        chunksize=CHUNKSIZE,
        usecols=[0, 1, 2, 3, 5],
    )
    return df


class AntibodyRegistryGetter(Obo):
    """An ontology representation of the Antibody Registry."""

    ontology = bioversions_key = PREFIX

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(force=force)


def get_obo(*, force: bool = False) -> Obo:
    """Get the Antibody Registry as OBO."""
    return AntibodyRegistryGetter(force=force)


# TODO there are tonnnnsss of mappings to be curated
MAPPING: Mapping[str, Optional[str]] = {
    "AMERICAN DIAGNOSTICA": None,  # No website
    "Biolegend": "biolegend",
    "Enzo Life Sciences": "enzo",
    "Novus": "novus",
    "LifeSpan": "biozil",
    "Creative Diagnostics": None,  # This site doesn't have a provider for IDs
}

SKIP = {
    "Universi",
    "School",
    "201",
    "200",
    "199",
}


def iter_terms(force: bool = False) -> Iterable[Term]:
    """Iterate over antibodies."""
    chunks = get_chunks(force=force)
    needs_curating = set()
    # df['vendor'] = df['vendor'].map(bioregistry.normalize_prefix)
    it = tqdm(chunks, desc=f"{PREFIX}, chunkssize={CHUNKSIZE}")
    for chunk in it:
        for identifier, name, vendor, catalog_number, defining_citation in chunk.values:
            if pd.isna(identifier):
                continue
            identifier = removeprefix(identifier, "AB_")
            term = Term.from_triple(PREFIX, identifier, name if pd.notna(name) else None)
            if vendor not in MAPPING:
                if vendor not in needs_curating:
                    needs_curating.add(vendor)
                    if all(x not in vendor for x in SKIP):
                        logger.debug(f"! vendor {vendor} for {identifier}")
            elif MAPPING[vendor] is not None and pd.notna(catalog_number) and catalog_number:
                term.append_xref((MAPPING[vendor], catalog_number))  # type:ignore
            if defining_citation and pd.notna(defining_citation):
                for pubmed_id in defining_citation.split(","):
                    pubmed_id = pubmed_id.strip()
                    if not pubmed_id:
                        continue
                    term.append_provenance(("pubmed", pubmed_id))
            yield term


if __name__ == "__main__":
    AntibodyRegistryGetter.cli()



# -*- coding: utf-8 -*-

"""Utilities for data structures for OBO."""

__all__ = [
    "OBO_ESCAPE_SLIM",
    "OBO_ESCAPE",
    "obo_escape",
    "obo_escape_slim",
    "comma_separate",
]

OBO_ESCAPE_SLIM = {c: f"\\{c}" for c in ':,"\\()[]{}'}
OBO_ESCAPE = {" ": "\\W", **OBO_ESCAPE_SLIM}


def obo_escape(string: str) -> str:
    """Escape all funny characters for OBO."""
    return "".join(OBO_ESCAPE.get(character, character) for character in string)


def obo_escape_slim(string: str) -> str:
    """Escape all funny characters for OBO."""
    rv = "".join(OBO_ESCAPE_SLIM.get(character, character) for character in string)
    rv = rv.replace("\n", "\\n")
    return rv


def comma_separate(elements) -> str:
    """Map a list to strings and make comma separated."""
    return ", ".join(map(str, elements))



# -*- coding: utf-8 -*-

"""Converter for WikiPathways."""

import logging
import urllib.error
from typing import Iterable

from .gmt_utils import parse_wikipathways_gmt
from ..constants import SPECIES_REMAPPING
from ..struct import Obo, Reference, Term, from_species
from ..struct.typedef import has_participant
from ..utils.path import ensure_path

__all__ = [
    "WikiPathwaysGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "wikipathways"

_PATHWAY_INFO = [
    ("Anopheles_gambiae", "7165"),
    ("Arabidopsis_thaliana", "3702"),
    ("Bos_taurus", "9913"),
    ("Caenorhabditis_elegans", "6239"),
    ("Canis_familiaris", "9615"),
    ("Danio_rerio", "7955"),
    ("Drosophila_melanogaster", "7227"),
    ("Equus_caballus", "9796"),
    ("Gallus_gallus", "9031"),
    ("Homo_sapiens", "9606"),
    ("Mus_musculus", "10090"),
    ("Oryza_sativa", "4530"),
    ("Pan_troglodytes", "9598"),
    ("Populus_trichocarpa", "3694"),
    ("Rattus_norvegicus", "10116"),
    ("Saccharomyces_cerevisiae", "4932"),
    ("Sus_scrofa", "9823"),
    ("Solanum_lycopersicum", "4081"),
]


class WikiPathwaysGetter(Obo):
    """An ontology representation of WikiPathways' pathway database."""

    ontology = bioversions_key = PREFIX
    typedefs = [from_species, has_participant]

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise)


def get_obo() -> Obo:
    """Get WikiPathways as OBO."""
    return WikiPathwaysGetter()


def iter_terms(version: str) -> Iterable[Term]:
    """Get WikiPathways terms."""
    base_url = f"http://data.wikipathways.org/{version}/gmt/wikipathways-{version}-gmt"

    for species_code, taxonomy_id in _PATHWAY_INFO:
        url = f"{base_url}-{species_code}.gmt"
        try:
            path = ensure_path(PREFIX, url=url, version=version)
        except urllib.error.HTTPError as e:
            logger.warning("HTTP %s error when downloading WikiPathways file %s", e.code, url)
            continue
        species_code = species_code.replace("_", " ")
        taxonomy_name = SPECIES_REMAPPING.get(species_code, species_code)

        for identifier, _version, _revision, name, _species, genes in parse_wikipathways_gmt(path):
            term = Term(reference=Reference(prefix=PREFIX, identifier=identifier, name=name))
            term.set_species(taxonomy_id, taxonomy_name)
            for ncbigene_id in genes:
                term.append_relationship(
                    has_participant,
                    Reference(prefix="ncbigene", identifier=ncbigene_id),
                )
            yield term


if __name__ == "__main__":
    WikiPathwaysGetter.cli()



# -*- coding: utf-8 -*-

"""Convert DrugBank Salts to OBO.

Run with ``python -m pyobo.sources.drugbank_salt``

Get relations between drugbank salts and drugbank parents with
``pyobo relations drugbank --relation obo:has_salt`` or

.. code-block:: python

    import pyobo
    df = pyobo.get_filtered_relations_df('drugbank', 'obo:has_salt')
"""

import logging
from typing import Iterable

from .drugbank import iterate_drug_info
from ..struct import Obo, Reference, Term

__all__ = [
    "DrugBankSaltGetter",
]

logger = logging.getLogger(__name__)

PREFIX = "drugbank.salt"


class DrugBankSaltGetter(Obo):
    """A getter for DrugBank Salts."""

    ontology = PREFIX
    bioversions_key = "drugbank"

    def iter_terms(self, force: bool = False) -> Iterable[Term]:
        """Iterate over terms in the ontology."""
        return iter_terms(version=self._version_or_raise, force=force)


def get_obo(force: bool = False) -> Obo:
    """Get DrugBank Salts as OBO."""
    return DrugBankSaltGetter(force=force)


def iter_terms(version: str, force: bool = False) -> Iterable[Term]:
    """Iterate over DrugBank Salt terms in OBO."""
    for drug_info in iterate_drug_info(version, force=force):
        for salt in drug_info.get("salts", []):
            xrefs = []
            for key in ["unii", "cas", "inchikey"]:
                identifier = salt.get(key)
                if identifier:
                    xrefs.append(Reference(prefix=key, identifier=identifier))

            yield Term(
                reference=Reference(
                    prefix=PREFIX,
                    identifier=salt["identifier"],
                    name=salt["name"],
                ),
                xrefs=xrefs,
            )


if __name__ == "__main__":
    DrugBankSaltGetter.cli()



# -*- coding: utf-8 -*-

"""Tools for loading entry points."""

from functools import lru_cache
from typing import Callable, Iterable, Mapping, Optional

from .struct import Obo

__all__ = [
    "has_nomenclature_plugin",
    "run_nomenclature_plugin",
    "iter_nomenclature_plugins",
]


@lru_cache()
def _get_nomenclature_plugins() -> Mapping[str, Callable[[], Obo]]:
    from .sources import ontology_resolver

    return ontology_resolver.lookup_dict


def has_nomenclature_plugin(prefix: str) -> bool:
    """Check if there's a plugin for converting the prefix."""
    from .sources import ontology_resolver

    try:
        ontology_resolver.lookup(prefix)
    except (KeyError, ValueError, TypeError):
        return False
    else:
        return True


def run_nomenclature_plugin(prefix: str, version: Optional[str] = None) -> Obo:
    """Get a converted PyOBO source."""
    from .sources import ontology_resolver

    return ontology_resolver.make(prefix, data_version=version)


def iter_nomenclature_plugins() -> Iterable[Obo]:
    """Get all modules in the PyOBO sources."""
    from .sources import ontology_resolver

    for _prefix, cls in sorted(ontology_resolver.lookup_dict.items()):
        yield cls()
